import os,sys,glob,re
import traceback
from ..resultQe import ResultQe
from .. import comm
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element
import numpy as np

xfmlt = comm.XmlFindMultiLayerText
xfml = comm.XmlFindMultiLayer

class Qe(ResultQe):
    @ResultQe.register(normal_end="if the job is normal ending")
    def GetNormalEnd(self):
        if self.XMLROOT:
            exit_status = self.XMLROOT.find('exit_status')
            if exit_status != None:
                if exit_status.text == '0':
                    self["normal_end"] = True
                else:
                    self["normal_end"] = False
                return   
            self["normal_end"] = None
        elif self.OUTPUT:
            if len(self.OUTPUT) > 1 and "JOB DONE." in self.OUTPUT[-2]:
                self["normal_end"] = True
            else:
                self["normal_end"] = False
        else:
            self["normal_end"] = None
            
    @ResultQe.register(version="the version of QE",
                       ncore = "the mpi cores")
    def GetGeneralInfo(self):
        if self.XMLROOT != None:
            self['version'] = self.XMLROOT.find("general_info/creator").attrib['VERSION']
            self['ncore'] = self.XMLROOT.find("parallel_info/nprocs").text
        elif self.OUTPUT:
            for iline in self.OUTPUT:
                if "Program PWSCF" in iline:
                    self['version'] = iline.split()[2].strip().lstrip("v.")
                elif "Parallel version (MPI)" in iline:
                    try:
                        self['ncore'] = int(iline.split()[5].strip())
                    except:
                        traceback.print_exc()
                        self['ncore'] = None
                    break
        else:
            self['version'] = None
            self['ncore'] = None
    
    @ResultQe.register(natom="total atom number",
                       ecutwfc="energy cutoff in Ry",
                       ks_solver="the diagonalization method",
                       mixing_type="the charge mixing method in SCF",
                       mixing_beta="the coefficient beta of charge mixing in SCF",
                       scf_thr="the convergence threshold of SCF, unit in Ry")
    def GetInputParam(self):
        if self.XMLROOT != None:

            input_param = self.XMLROOT.find('input')        

            self['natom'] = comm.iint(input_param.find('atomic_structure').attrib['nat'])


            ecutwfc = comm.ifloat(xfmlt(input_param,['basis','ecutwfc']))
            if ecutwfc != None: ecutwfc *= 2
            self["ecutwfc"] = ecutwfc

            self['ks_solver'] = xfmlt(input_param,['electron_control','diagonalization'])
            self['mixing_type'] = xfmlt(input_param,['electron_control','mixing_mode'])
            self['mixing_beta'] = comm.ifloat(xfmlt(input_param,['electron_control','mixing_beta']))

            scf_thr = comm.ifloat(xfmlt(input_param,['electron_control','conv_thr']))
            if scf_thr != None: scf_thr *= 2
            self['scf_thr'] = scf_thr
        elif self.OUTPUT:
            ks_solver = None
            for iline in self.OUTPUT:
                if "number of atoms/cell" in iline:
                    self["natom"] = int(iline.split()[4])
                elif "kinetic-energy cutoff" in iline:
                    self["ecutwfc"] = float(iline.split()[4])
                elif "diagonalization" in iline:
                    ks_solver = iline.split()[0].lower()
                elif "mixing beta" in iline:
                    self["mixing_beta"] = float(iline.split()[3])
                elif "number of iterations used =" in iline:
                    self["mixing_type"] = iline.split()[6]
                elif "convergence threshold" in iline:
                    self["scf_thr"] = float(iline.split()[4])
            self["ks_solver"] = ks_solver
        else:
            self["natom"] = None
            self["ecutwfc"] = None
            self["ks_solver"] = None
            self["mixing_type"] = None
            self["mixing_beta"] = None
            self["scf_thr"] = None

    @ResultQe.register(smearing_method="smearing method",
                       smearing_sigma = "smearing sigma in Ry")
    def GetSmearing(self):
        if self.XMLROOT != None:
            input_param = self.XMLROOT.find('input')

            smearing_method = xfmlt(input_param,['bands','occupations'])
            if smearing_method == 'fixed':
                smearing_sigma = 0.0
            elif smearing_method == 'smearing':
                smearing_method = xfmlt(input_param,['bands','smearing'])
                smearing_sigma = comm.ifloat(xfml(input_param,['bands','smearing']).attrib.get('degauss',0.0))
                if smearing_sigma != None:
                    smearing_sigma *= 2.0

            self["smearing_method"] = smearing_method
            self["smearing_sigma"] = smearing_sigma
        else:
            self["smearing_method"] = None
            self["smearing_sigma"] = None
            
    
    @ResultQe.register(converge="if SCF is converged",
                       scf_steps="the steps of SCF",
                       energy="total energy, unit in eV",
                       energies = "list, the total energy of each ION step",
                       nbands="band number",
                       ibzk="irreducible K point number",
                       nkstot="the total k points",
                       kpt="list, the K POINTS setting",
                       force="list, the force of all atoms, [atom1x,atom1y,atom1z,atom2x,atom2y,atom2z...]. Unit in eV/Angstrom",
                       stress="list, the stress, [xx,xy,xz,yx,yy,yz,zx,zy,zz]. Unit in kbar.",
                       virial="list, the virial, [xx,xy,xz,yx,yy,yz,zx,zy,zz]. Unit in eV.",
                       pressure="the pressure, unit in kbar",
                       forces = "list, the force of each ION step",
                       stresses = "list, the stress of each ION step",
                       virials = "list, the virial of each ION step",
                       pressures = "list, the pressure of each ION step",
                       cell = "list, the cell, [a1,a2,a3,b1,b2,b3,c1,c2,c3]. Unit in Angstrom",
                       volume = "the volume of cell, unit in Angstrom^3",
                       coord = "list, the coordinate of all atoms, [atom1x,atom1y,atom1z,atom2x,atom2y,atom2z...]. Unit in Angstrom",
                       label = "the label of each atom",
                       atomlabel_list = "same as label",
                       element = "the element of each atom. Because QE do not output the element, we actually output the label.",
                       element_list = "same as element.",
                       total_mag="total magnization",
                       absolute_mag="total absolute magnization",
                       atom_mag = "list, the magnization of each atom",
                       nelec="total electron number",
                       energy_per_atom="total energy divided by natom, unit in eV",)
    def GetOutputParam(self):
        output = None
        if self.XMLROOT != None:
            output = self.XMLROOT.findall('output')
            if len(output) == 0:
                print("ERROR: can not find the output in pwscf.xml")
                return
            output = output[-1]

            element = None
            self['converge'] = comm.ibool(xfmlt(output,['convergence_info','scf_conv','convergence_achieved']))
            self['scf_steps'] = comm.iint(xfmlt(output,['convergence_info','scf_conv','n_scf_steps']))
            self['ibzk'] = comm.iint(xfmlt(output,['band_structure','nks']))
            self['nelec'] = comm.ifloat(xfmlt(output,['band_structure','nelec']))
            self['nbands'] = comm.iint(xfmlt(output,['band_structure','nbnd']))

            total_energy = comm.ifloat(xfmlt(output,['total_energy','etot']))
            if total_energy != None: total_energy *= comm.HARTREE2EV
            self['energy'] = total_energy

            mp = xfml(output,['band_structure','starting_k_points','monkhorst_pack'])
            if isinstance(mp, Element):
                nk1 = comm.iint(mp.get('nk1'))
                nk2 = comm.iint(mp.get('nk2'))
                nk3 = comm.iint(mp.get('nk3'))
                if nk1 and nk2 and nk3: 
                    self['nkstot'] = nk1 * nk2 * nk3
                    self['kpt'] = [nk1, nk2, nk3]

            # structure
            structure = output.findall('atomic_structure')
            if len(structure) == 0:
                structure = None
            else:
                structure = structure[-1]
            volume = None
            if structure != None:
                cell_a = xfmlt(structure,['cell','a1'])
                cell_b = xfmlt(structure,['cell','a2'])
                cell_c = xfmlt(structure,['cell','a3'])
                if cell_a and cell_b and cell_c:
                    self['cell'] = cell = [[float(i)* comm.BOHR2A for i in j.split()]  for j in [cell_a,cell_b,cell_c]]
                    # calculate the volume by cell
                    
                    self["volume"] = volume = np.linalg.det(cell)
                coord = []
                label = []
                for icoord in structure.findall('atomic_positions/atom'):
                    coord.append([float(i) * comm.BOHR2A for i in icoord.text.split()])
                    label.append(icoord.attrib['name'])
                if len(coord) == self['natom']:
                    self['coord'] = coord
                    self["label"] = label
                    element = label
                else:
                    self['coord'] = None
                    self["label"] = None
                    element = None
                    print("ERROR: the length of coord is not equal to natom")
                self["atomlabel_list"] = self["label"]
            else:
                self['cell'] = None
                self['coord'] = None

            if output.find('forces') != None:
                self['force'] = [float(i) * comm.HARTREE2EV / comm.BOHR2A for i in output.find('forces').text.split()]
            if output.find('stress') != None:
                # the read in stress is in Hartree/Bohr^3, convert to kbar
                # 1 kbar = 1e8 Pa = 1e8 N/m^2 = 1e8 J/m^3 = 1e8 * 2.2937126583579E17 Hartree/m^3 = 2.2937126583579E25 * 5.29177E-11**3 Hartree/Bohr^3 = 3.398927420868445E-6 Hartree/Bohr^3
                self['stress'] = [ float(i)/comm.KBAR2HARTREEPERBOHR3 for i in output.find('stress').text.split()]
                self["pressure"] = (self['stress'][0] + self['stress'][4] + self['stress'][8]) / 3.0
                # calculate the virial, unit in eV
                if volume != None:
                    self['virial'] = [ float(i) * volume * comm.HARTREE2EV / comm.BOHR2A ** 3   for i in output.find('stress').text.split()]

            tot_mag = comm.ifloat(xfmlt(output,['magnetization','total']))
            if tot_mag == None and output.find('magnetization/total_vec') != None:
                tot_mag = [float(i) for i in output.find('magnetization/total_vec').text.split()]
            self['total_mag'] = tot_mag if tot_mag != None else None
            self['absolute_mag'] = comm.ifloat(xfmlt(output,['magnetization','absolute']))
            atom_mags = output.findall('magnetization/Scalar_Site_Magnetic_Moments/SiteMagnetization')
            if atom_mags:
                atom_mag = []
                element = []
                for i in atom_mags:
                    atom_mag.append(float(i.text))
                    element.append(i.attrib['species'].strip())
                self['atom_mag'] = atom_mag
            else:
                self['atom_mag'] = None
                
            if self['energy'] == None or self['natom'] == None:
                self['energy_per_atom'] = None
            else:
                self['energy_per_atom'] = self['energy'] / self['natom']
            self["element"] = element
            self["element_list"] = element
            
            allsteps = self.XMLROOT.findall('step')
            energies = []
            forces = []
            stresses = []
            pressures = []
            virials = []
            for istep in allsteps:
                energies.append(comm.ifloat(xfmlt(istep,['total_energy','etot'])) * comm.HARTREE2EV)
                force = istep.find('forces')
                if force != None:
                    forces.append([float(i) * comm.HARTREE2EV / comm.BOHR2A for i in force.text.split()])
                stress = istep.find('stress')
                if stress != None:
                    stresses.append([float(i)/comm.KBAR2HARTREEPERBOHR3 for i in stress.text.split()])
                    pressures.append((stresses[-1][0] + stresses[-1][4] + stresses[-1][8]) / 3.0)
                    if volume != None:
                        virials.append([float(i) * volume * comm.HARTREE2EV / comm.BOHR2A ** 3   for i in stress.text.split()])
            self["energies"] = energies if energies else None
            self["forces"] = forces if forces else None
            self["stresses"] = stresses if stresses else None
            self["pressures"] = pressures if pressures else None
            self["virials"] = virials if virials else None
            
        elif self.OUTPUT:
            converge = None
            ibzk = None
            nbands = None
            nelec = None
            energies = []
            stresses = []
            pressures = []
            forces = []
            natom = None
            cell = None
            coord = None
            label = None
            element = None
            tot_mag = None
            abs_mag = None
            atom_mag = None
            scf_steps = None
            for idx,iline in enumerate(self.OUTPUT):
                if "convergence has been achieved" in iline:
                    converge = True
                elif "convergence NOT achieved" in iline:
                    converge = False
                elif "atomic species   valence    mass     pseudopotential" in iline:
                    j = idx + 1
                    element = {}
                    while self.OUTPUT[j].strip() !="":
                        species = self.OUTPUT[j].split()[0]
                        element[species] = self.OUTPUT[j].split("(")[0].split()[-1].strip()
                        j += 1
                elif "number of k points" in iline:
                    ibzk = int(iline.split()[4])
                elif "total energy" in iline:
                    energies.append(float(iline.split()[-2]) * comm.RY2EV)
                elif "number of Kohn-Sham states" in iline:
                    nbands = int(iline.split()[-1])
                elif "number of electrons" in iline:
                    nelec = float(iline.split()[-1])
                elif "number of atoms/cell" in iline:
                    natom = int(iline.split()[-1])
                elif "total magnetization" in iline:
                    sline = iline.split()
                    if len(sline) == 8:
                        tot_mag = [float(imag) for imag in sline[3:6]]
                    else:
                        tot_mag = float(sline[3])
                elif "absolute magnetization" in iline:
                    abs_mag = float(iline.split()[3])
                elif "iteration #" in iline:
                    if scf_steps == None:
                        scf_steps = 1
                    else:
                        scf_steps += 1
                elif "magn=" in iline:
                    if atom_mag == None:
                        atom_mag = []
                    atom_mag.append(float(iline.split()[-1]))
                elif "total stress" in iline:
                    stress = []
                    for i in range(3):
                        stress.extend([float(j) for j in self.OUTPUT[idx+i+1].split()[3:6]])
                    pressure = float(iline.split()[-1])
                    stresses.append(stress)
                    pressures.append(pressure)
                elif "Forces acting on atoms" in iline:
                    force = []
                    j = idx + 2
                    while len(self.OUTPUT[j].split()) == 8:
                        force.extend([float(i) * comm.RY2EV / comm.BOHR2A for i in self.OUTPUT[j].split()[-3:]])
                        j += 1
                    forces.append(force)
                elif "CELL_PARAMETERS" in iline:
                    cell = []
                    lat0 = float(iline.split("=")[1].strip()[:-1])
                    for i in range(3):
                        cell.append([float(j) * lat0 for j in self.OUTPUT[idx+i+1].split()])
                elif "ATOMIC_POSITIONS" in iline:
                    coord = []
                    label = []
                    if natom == None:
                        print("ERROR: can not find natom when catch coord")
                        continue
                    for i in range(natom):
                        icoord = self.OUTPUT[self.OUTPUT.index(iline)+i+1].split()
                        label.append(icoord[0])
                        coord.append([float(j) for j in icoord[1:]])
                elif "site n.     atom                  positions (alat units)" in iline:
                    label = []
                    j = idx + 1
                    while self.OUTPUT[j].strip() != "":
                        label.append(self.OUTPUT[j].split()[1])
                        j += 1
                        
            self["converge"] = converge
            self["energy"] = None if len(energies) == 0 else energies[-1]
            self["energies"] = None if len(energies) == 0 else energies
            self["nbands"] = nbands
            self["ibzk"] = ibzk
            self["nelec"] = nelec
            self["stress"] = None if len(stresses) == 0 else stresses[-1]
            self["stresses"] = None if len(stresses) == 0 else stresses
            self["pressure"] = None if len(pressures) == 0 else pressures[-1]
            self["pressures"] = None if len(pressures) == 0 else pressures
            self["force"] = None if len(forces) == 0 else forces[-1]
            self["forces"] = None if len(forces) == 0 else forces
            self["natom"] = natom
            self["cell"] = cell
            self["coord"] = coord
            self["label"] = label
            self["atomlabel_list"] = label
            self["element"] = None if label ==None or element == None else [element[i] for i in label]
            self["element_list"] = label
            self["total_mag"] = tot_mag
            self["absolute_mag"] = abs_mag
            self["atom_mag"] = atom_mag
            self["scf_steps"] = scf_steps
            self["nkstot"] = None
            self["kpt"] = None
            if energies and natom != None:
                self["energy_per_atom"] = energies[-1] / natom
            else:
                self["energy_per_atom"] = None
            
            if cell != None:
                self["volume"] = abs(np.linalg.det(cell))
            else:
                self["volume"] = None
            
            # calc the virial
            if stresses and self["volume"] != None:
                virials = []
                for stress in stresses:
                    virials.append([i * self["volume"] * comm.KBAR2EVPERANGSTROM3 for i in stress])
                self["virial"] = virials[-1]
                self["virials"] = virials
            else:
                self["virial"] = None
                self["virials"] = None
        else:
            self["converge"] = None
            self["scf_steps"] = None
            self["energy"] = None
            self["energies"] = None
            self["nbands"] = None
            self["ibzk"] = None
            self["nkstot"] = None
            self["kpt"] = None
            self["force"] = None
            self["forces"] = None
            self["stress"] = None
            self["stresses"] = None
            self["virial"] = None
            self["virials"] = None
            self["pressure"] = None
            self["pressures"] = None
            self["cell"] = None
            self["volume"] = None
            self["coord"] = None
            self["label"] = None
            self["atomlabel_list"] = None
            self["element"] = None
            self["element_list"] = None
            self["total_mag"] = None
            self["absolute_mag"] = None
            self["atom_mag"] = None
            self["nelec"] = None
            self["energy_per_atom"] = None
            
    @ResultQe.register(band="list, the band, [[spin1],[spin2],...], unit in eV",
                       efermi="the Fermi energy, unit in eV",
                       band_gap="the band gap, unit in eV")
    def GetBand(self):  
        band = None
        efermi = None
        band_gap = None
        if self.XMLROOT != None:
            band_all = []
            for ik in self.XMLROOT.findall('output/band_structure/ks_energies/eigenvalues'):
                band_all.append([float(i) * comm.HARTREE2EV for i in ik.text.split()])
            nbnd_up = comm.iint(xfmlt(self.XMLROOT,['output','band_structure','nbnd_up']))
            if nbnd_up != None:
                band = [[],[]]
                for ik in band_all:
                    band[0].append(ik[:nbnd_up])
                    band[1].append(ik[nbnd_up:])
            else:
                band = [band_all]     
                
            efermi = comm.ifloat(xfmlt(self.XMLROOT,['output','band_structure','fermi_energy']))
            if efermi != None:
                efermi *= comm.HARTREE2EV
        elif self.OUTPUT:
            band_all = []
            for idx,iline in enumerate(self.OUTPUT):
                if "the Fermi energy is" in iline:
                    efermi = float(iline.split()[-2])
                    break
                elif " k =" in iline:
                    j = idx + 2
                    band_all.append([])
                    while self.OUTPUT[j].strip() != "":
                        band_all[-1].extend([float(i) for i in self.OUTPUT[j].split()])
                        j += 1
            ibzk = self["ibzk"]
            if ibzk != None and len(band_all) == 2*ibzk:
                band = [band_all[:ibzk],band_all[ibzk:]]
            elif band:
                band = [band_all]
            
        if band != None and efermi != None:
            band_gap = comm.cal_band_gap(band,efermi)
        self["band"] = band
        self["efermi"] = efermi
        self["band_gap"] = band_gap      
    
    @ResultQe.register(drho="[], drho of each scf step. Unit in Ry",
                       drho_last="drho of the last scf step. Unit in Ry",
                       denergy="[], denergy of each scf step. Unit in eV",
                       denergy_last="denergy of the last scf step. Unit in eV")
    def Getdrhode(self):
        drho = None
        drho_last = None
        denergy = None
        denergy_last = None
        energy = None
        if self.OUTPUT:
            for iline in self.OUTPUT:
                if "total energy              =" in iline:
                    if energy == None:
                        energy = []
                    energy.append(iline.split()[-2])
                elif "estimated scf accuracy    <" in iline:
                    if drho == None:
                        drho = []
                    drho.append(float(iline.split()[-2]))
        
        if drho != None:
            drho_last = drho[-1]
        if energy != None:
            e_acc = 10**(-1 * len(energy[0].split(".")[1]))
            de = e_acc* comm.RY2EV
            denergy = [0.0]
            for i in range(1,len(energy)):
                e1 = float(energy[i])
                e2 = float(energy[i-1])
                if not (abs(e1 - e2) < e_acc):
                    de = (e1 - e2)*comm.RY2EV
                denergy.append(de)
            denergy_last = denergy[-1]
            
            
        self["drho"] = drho
        self["drho_last"] = drho_last
        self["denergy"] = denergy
        self["denergy_last"] = denergy_last
                    
            
    
    @ResultQe.register(total_time="the total running time (WALL)",
                       force_time="the time to do the calculations of force (WALL)",
                       stress_time="the time to do the calculation of stress (WALL)")
    def GetTimeParam(self):
        if self.XMLROOT != None:

            timing = self.XMLROOT.find('timing_info')

            self['total_time'] = comm.ifloat(xfmlt(timing,['total','wall']))

            all_time = timing.findall('partial')
            for itime in all_time:
                if itime.get('label') == 'forces':
                    self['force_time'] = comm.ifloat(itime.find('wall').text)
                elif itime.get('label') == 'stress':
                    self['stress_time'] = comm.ifloat(itime.find('wall').text)
        elif self.OUTPUT:
            # read the time from the output file
            # begin to read from the end of the file
            force_time = None
            stress_time = None
            total_time = None
            
            for iline in self.OUTPUT[::-1]:
                if "Writing all to output data dir" in iline:
                    break
                if "PWSCF        :" in iline:
                    # may have case "PWSCF        :      7h43m CPU      8h 0m WALL"
                    total_time = comm.strtime2sec(iline.split("CPU")[-1].split("WALL")[0])
                elif "force" in iline:
                    force_time = comm.strtime2sec(iline.split()[-2])
                elif "stress" in iline:
                    stress_time = comm.strtime2sec(iline.split()[-2])
            self["total_time"] = total_time
            self["force_time"] = force_time
            self["stress_time"] = stress_time
        else:
            self["total_time"] = None
            self["force_time"] = None
            self["stress_time"] = None

    @ResultQe.register(scf_time="the SCF time (WALL)",
                       scf_time_each_step="list, the SCF time of each step (WALL)",
                       scf_time_step1 = "the time of the first step (WALL)")
    def GetSCFTime(self):
        if self.OUTPUT:
            start_time = None # the time before SCF
            scf_times = [] # the time of each SCF step
            scf_loop = False
            for i,line in enumerate(self.OUTPUT):
                if "Self-consistent Calculation" in line:
                    for j in range(i-1,-1,-1):
                        if "total cpu time spent up to now is" in self.OUTPUT[j]:
                            start_time = float(self.OUTPUT[j].split()[-2])
                            break
                    scf_loop = True
                elif "End of self-consistent calculation" in line:
                    break
                elif "total cpu time spent up to now is" in line and scf_loop and start_time != None:
                    scf_times.append(float(line.split()[-2]) - start_time)
                    start_time = float(line.split()[-2])
            self["scf_time"] = None if not scf_times else sum(scf_times)
            self["scf_time_each_step"] = None if not scf_times else scf_times
            self["scf_time_step1"] = None if not scf_times else scf_times[0]
        else:
            self["scf_time"] = None
            self["scf_time_each_step"] = None
            self["scf_time_step1"] = None
            
                    
                    
    
    @ResultQe.register(relax_converge="if the relax is converged")
    def GetRelaxConverge(self):
        if self.OUTPUT:
            for line in self.OUTPUT[::-1]:
                if "The maximum number of steps has been reached." in line:
                    self["relax_converge"] = False
                    return
                elif "bfgs converged in" in line:
                    self["relax_converge"] = True
                    return
            self["relax_converge"] = None
            
        elif self.XMLROOT != None:
            output = self.XMLROOT.find('output')
            rlx_conv = comm.ibool(xfmlt(output,['convergence_info','opt_conv','convergence_achieved']))
            # in QE, after the rlx is converged, it will do another SCF, but in some case the scf is not done, and this value will be false
            # we need to check if the ION steps is less than input/control_variables/nstep, if yes, we should set this value to True
            if not rlx_conv:
                nstep = comm.iint(xfmlt(self.XMLROOT,["input",'control_variables','nstep']))
                if nstep == None:
                    nstep = 50
                if self["relax_steps"] != None and self["relax_steps"] < nstep:
                    rlx_conv = True
            
            self["relax_converge"] = rlx_conv
        else:
            self["relax_converge"] = None

    @ResultQe.register(relax_steps="the total ION steps")
    def GetRelaxStepsFromXml(self):
        if self.XMLROOT != None:
            self["relax_steps"] = len(self.XMLROOT.findall('step'))
        elif self.OUTPUT:
            for iline in self.OUTPUT[::-1]:
                if "number of bfgs steps" in iline:
                    self["relax_steps"] = int(iline.split()[-1])
                    return
                
        self["relax_steps"] = None
        
        




"""Module for managing DOS and PDOS data from ABACUS calculations."""

import numpy as np
from typing import Dict, List, Tuple, Optional, Union

from abacustest.lib_collectdata.collectdata import RESULT


l_map = ['s', 'p', 'd', 'f', 'g']

orbital_names = {
    (0, 0): "s",
    (1, 0): r"$p_z$",
    (1, 1): r"$p_x$",
    (1, 2): r"$p_y$",
    (2, 0): r"$d_{z^2}$",
    (2, 1): r"$d_{xz}$",
    (2, 2): r"$d_{yz}$",
    (2, 3): r"$d_{x^2-y^2}$",
    (2, 4): r"$d_{xy}$",
    (3, 0): r"$f_{z^3}$",
    (3, 1): r"$f_{xz^2}$",
    (3, 2): r"$f_{yz^2}$",
    (3, 3): r"$f_{zx^2-zy^2}$",
    (3, 4): r"$f_{xyz}$",
    (3, 5): r"$f_{x^3-3xy^2}$",
    (3, 6): r"$f_{3yx^2-y^3}$",
    (4, 0): r"$g_1$",
    (4, 1): r"$g_2$",
    (4, 2): r"$g_3$",
    (4, 3): r"$g_4$",
    (4, 4): r"$g_5$",
    (4, 5): r"$g_6$",
    (4, 6): r"$g_7$",
    (4, 7): r"$g_8$",
    (4, 8): r"$g_9$",
}

color_list = [
    '#1f77b4',  # 蓝色
    '#ff7f0e',  # 橙色
    '#2ca02c',  # 绿色
    '#d62728',  # 红色
    '#9467bd',  # 紫色
    '#8c564b',  # 棕色
    '#e377c2',  # 粉色
    '#7f7f7f',  # 灰色
    '#bcbd22',  # 黄绿色
    '#17becf',  # 青色
    '#aec7e8',  # 浅蓝色
    '#ffbb78',  # 浅橙色
    '#98df8a',  # 浅绿色
    '#ff9896',  # 浅红色
]

class DOSData:
    """Class for managing DOS data from ABACUS calculations."""
    def __init__(self,
                 energy: np.ndarray = None,
                 dosdata: np.ndarray = None,
                 efermi: float=None):
        """
        Args:
            energy: Original energy values for the DOS.
            dosdata: Total DOS data. If nspin=1 or 4, should be a 1D array with shape (nenergy,), or nspin=2 with shape (nenergy, 2).
            efermi: Fermi energy.
        """
        self.efermi = efermi
        if efermi is not None:
            self.energy = energy - efermi
        else:
            self.energy = energy
        
        assert energy.shape[0] == dosdata.shape[0]

        self.dosdata = dosdata
    
    @staticmethod
    def ReadFromAbacusJob(abacusjob_dir: str, efermi: Optional[float] = None) -> 'DOSData':
        """
        Read DOS outputed by ABACUS from directory of finished ABACUS calculation.
        """
        results = RESULT(fmt="abacus", path=abacusjob_dir)
        energy = np.array(results['dos']['energy'])
        dosdata = np.array(results['dos']['data'])
        if efermi is None:
            efermi = results['efermi']

        return DOSData(energy=energy, dosdata=dosdata, efermi=efermi)
    
    def plot_dos(self, emin: float, emax: float, title: str, fname: str):
        """
        Plot DOS using the given energy and dosdata.
        """
        plot_dos_pdos([[self.dosdata]],
                      [['DOS']],
                      [title],
                      self.energy,
                      emin,
                      emax,
                      self.efermi is not None,
                      fname)
    
    def write_dos(self, fname: str):
        """
        Write DOS data to a file.
        """
        write_dos_pdos([self.dosdata],
                       self.energy,
                       ['DOS'],
                       self.efermi is not None,
                       fname)


class PDOSData:
    """Class for managing PDOS data from ABACUS calculations."""

    def __init__(self,
                 energy: np.ndarray = None, 
                 pdosdata: List[Dict[str, Tuple[int, np.ndarray]]] = None,
                 efermi: float=None):
        """
        Initialize DosObj to manage DOS and PDOS data from ABACUS output files.
        Parameters:
        -----------
        energy (np.ndarray):
            Original energy data outputed in DOS and PDOS calculation.
        pdosdata (List[Dict[str, Tuple[int, np.ndarray]]]):
            PDOS data for each orbital. Each element is a dictionary containing orbital information and PDOS data, with keys "atom_index", "l", "m", "z" and "data". The format of "data" is same with dosdata.
        efermi (float):
            Fermi energy. Original energy data is subtracted by this value. If None, will use the original energy data.
        """
        self.efermi = efermi
        if efermi is not None:
            self.energy = np.array(energy) - efermi

        for orbital_pdos in pdosdata:
            assert energy.shape[0] == orbital_pdos['data'].shape[0] # Length of energy and PDOS data of one orbital must match
            
        self.projected_dos = pdosdata

    def read_projected_dos(abacusjob_dir):
        """Read projected DOS from ABACUS output files (XML format)."""
        results = RESULT(fmt="abacus", path=abacusjob_dir)
        energy = np.array(results['pdos']['energy'])
        for i in range(len(results['pdos']['orbitals'])):
            results['pdos']['orbitals'][i]['data'] = np.array(results['pdos']['orbitals'][i]['data'])
        return energy, results['pdos']['orbitals']
    
    @staticmethod
    def ReadFromAbacusJob(abacus_job: str, efermi: Optional[float]=None) -> 'PDOSData':
        """
        Create a PDOSData object from an ABACUS job object.
        
        Parameters:
        -----------
        abacus_job : AbacusJob
            An instance of the AbacusJob class representing an ABACUS calculation.
        
        Returns:
        --------
        PDOSData
            An instance of the PDOSData class with data read from the ABACUS job output.
        """
        results = RESULT(fmt="abacus", path=abacus_job)
        energy, pdosdata = PDOSData.read_projected_dos(abacus_job)
        if efermi is None:
            efermi = results['efermi']

        return PDOSData(energy=energy, pdosdata=pdosdata, efermi=efermi)

    def get_pdos_by_orbital(self, species: str, atom_index: int, l: int, 
                            m: Optional[int] = None, z: Optional[int] = None) -> Dict:
        """
        Get PDOS data for a specific orbital.
        
        Parameters:
        -----------
        species : str
            Species name (e.g. 'Fe', 'O')
        atom_index : int
            Index of the atom (0-based)
        l : int
            Angular momentum quantum number (0=s, 1=p, 2=d, 3=f, ...)
        m : int, optional
            Magnetic quantum number
        z : int, optional
            Orbital split index
        
        Returns:
        --------
        dict or None
            Dictionary containing orbital information and PDOS data
        """
        for orbital in self.projected_dos:
            if (orbital['species'] == species and 
                orbital['atom_index'] == atom_index and 
                orbital['l'] == l):
                
                # If m and z are specified, match those as well
                if m is not None and orbital['m'] != m:
                    continue
                if z is not None and orbital['z'] != z:
                    continue
                
                return orbital
        
        return None
    
    def get_pdos_by_atom(self, atom_index: int, sum_only: bool=True) -> List[Dict]:
        """
        Get all PDOS data for a specific atom.
        
        Parameters:
        -----------
        atom_index : int
            Index of the atom (0-based)
        sum_only : bool, optional
            If True, sum the PDOS data for each orbital, otherwise return the original data. Default is True.
        
        Returns:
        --------
        list of dict
            List of orbitals belonging to the specified atom
        """
        pdos_datas = [orb for orb in self.projected_dos if orb['atom_index'] == atom_index]
        if sum_only:
            return PDOSData.sum_pdos_data(pdos_datas)
        else:
            return pdos_datas

    def get_pdos_by_atom_shell(self, atom_index: int, l: Tuple[int, str], sum_only: bool=True) -> List[Dict]:
        """
        Get PDOS data for a specific atom shell.
        
        Parameters:
        -----------
        atom_index : int
            Index of the atom (0-based)
        l : int or str
            Angular momentum quantum number (0 or s, 1 or p, 2 or d, 3 or f, ...)
        sum_only : bool, optional
            If True, sum the PDOS data for each orbital, otherwise return the original data. Default is True.
        
        Returns:
        --------
        list of dict
            Dictionary containing orbital information and PDOS data
        """
        if l in l_map:
            l = l_map.index(l)
        pdos_datas = [orb for orb in self.projected_dos if (orb['atom_index'] == atom_index and orb['l'] == l)]
        if sum_only:
            return PDOSData.sum_pdos_data(pdos_datas)
        else:
            return pdos_datas
    
    def get_pdos_by_atom_orbital(self, atom_index: int, l: Optional[Tuple[int, str]], m: Tuple[int, str], sum_only: bool=True) -> List[Dict]:
        """
        Get PDOS data for a specific atom orbital.
        
        Parameters:
        -----------
        atom_index : int
            Index of the atom (0-based)
        l : int or str
            Angular momentum quantum number (0 or s, 1 or p, 2 or d, 3 or f, ...). If m is explicitly specified by atomic orbital name, l is ignored.
        m : int or str
            Magnetic quantum number used by PDOS file in ABACUS.
            If m is an integer, it should be between 0 an 2l.
            If m is a string, it should be the name of the atomic orbital.
        sum_only : bool, optional
            If True, sum all the PDOS data for the specified atom orbital, otherwise return all the individual PDOS data.
        
        Returns:
        --------
        dict or None
            Dictionary containing orbital information and PDOS data
        """
        if isinstance(m, str):
            pdos_datas = [orb for orb in self.projected_dos if (orb['atom_index'] == atom_index and orb['atomic_orbital_name'] == m)]
        elif isinstance(m, int):
            if l in l_map:
                l = l_map.index(l)
            pdos_datas = [orb for orb in self.projected_dos if (orb['atom_index'] == atom_index and orb['l'] == l and orb['m'] == m)]
        else:
            raise ValueError("Invalid m value")
        
        if sum_only:
            return PDOSData.sum_pdos_data(pdos_datas)
        else:
            return pdos_datas

    def get_pdos_by_species(self, species: str, sum_only: bool=True) -> Union[np.ndarray, List[Dict]]:
        """
        Get PDOS data for a specific species.
    
        Parameters:
        -----------
        species : str
            Species name (e.g. 'Fe', 'O')
        sum_only : bool, optional
            If True, sum the PDOS data for each orbital, otherwise return the original data. Default is True.
    
        Returns:
        --------
        list of dict or np.ndarray
            If sum_only is True, returns summed PDOS data array with shape (nspin, len(energy)).
            Otherwise, returns list of orbitals belonging to the specified species
        """
        pdos_datas = [orb for orb in self.projected_dos if orb['species'] == species]
        if sum_only:
            return PDOSData.sum_pdos_data(pdos_datas)
        else:
            return pdos_datas

    def get_pdos_by_species_shell(self, species: str, l: Union[int, str], sum_only: bool=True) -> Union[np.ndarray, List[Dict]]:
        """
        Get PDOS data for shell of a specific species.
    
        Parameters:
        -----------
        species : str
            Species name (e.g. 'Fe', 'O')
        l : int or str
            Angular momentum quantum number (0 or 's', 1 or 'p', 2 or 'd', 3 or 'f', ...)
        sum_only : bool, optional
            If True, sum the PDOS data for each orbital, otherwise return the original data. Default is True.
    
        Returns:
        --------
        list of dict or np.ndarray
            If sum_only is True, returns summed PDOS data array with shape (nspin, len(energy))
            Otherwise, returns list of orbitals belonging to the specified species and shell
        """
        if l in l_map:
            l = l_map.index(l)
        pdos_datas = [orb for orb in self.projected_dos if (orb['species'] == species and orb['l'] == l)]
    
        if sum_only:
            return PDOSData.sum_pdos_data(pdos_datas)
        else:
            return pdos_datas
    
    def get_pdos_by_species_orbital(self, species: str, l: Optional[Union[int, str]], m: Union[int, str], sum_only: bool=True) -> Union[np.ndarray, List[Dict]]:
        """
        Get PDOS data for a specific species orbital.
    
        Parameters:
        -----------
        species : str
            Species name (e.g. 'Fe', 'O')
        l : int or str, optional
            Angular momentum quantum number (0 or 's', 1 or 'p', 2 or 'd', 3 or 'f', ...). 
            If m is explicitly specified by atomic orbital name, l is ignored.
        m : int or str
            Magnetic quantum number used by PDOS file in ABACUS.
            If m is an integer, it should be between 0 and 2l.
            If m is a string, it should be the name of the atomic orbital (e.g. 's', 'px', 'py', 'pz', 'dz2').
        sum_only : bool, optional
            If True, sum all the PDOS data for the specified species orbital, otherwise return all the individual PDOS data.
    
        Returns:
        --------
        list of dict or np.ndarray
            If sum_only is True, returns summed PDOS data array with shape (nspin, len(energy))
            Otherwise, returns list of orbitals belonging to the specified species and orbital
        """
        if isinstance(m, str):
            pdos_datas = [orb for orb in self.projected_dos if (orb['species'] == species and orb['atomic_orbital_name'] == m)]
        elif isinstance(m, int):
            if l in l_map:
                l = l_map.index(l)
            pdos_datas = [orb for orb in self.projected_dos if (orb['species'] == species and orb['l'] == l and orb['m'] == m)]
        else:
            raise ValueError("Invalid m value")
        
        if sum_only:
            return PDOSData.sum_pdos_data(pdos_datas)
        else:
            return pdos_datas
    
    def get_species(self) -> List[str]:
        # Get all species in the PDOS data
        return sorted(list(set([orb['species'] for orb in self.projected_dos])))
    
    def get_species_shell(self, species: str) -> List[int]:
        # Get all shells for a specific species in the PDOS data
        return list(set([orb['l'] for orb in self.projected_dos if orb['species'] == species]))
    
    def get_species_shell_orbital(self, species: str, l: int) -> List[str]:
        # Get all orbitals for a specific species and shell in the PDOS data
        return list(set([orb['m'] for orb in self.projected_dos if (orb['species'] == species and orb['l'] == l)]))
    
    def get_atom_species(self, atom_index: int) -> str:
        # Get the species for a specific atom in the PDOS data
        species = list(set([orb['species'] for orb in self.projected_dos if orb['atom_index'] == atom_index]))
        assert len(species) == 1
        return species[0]

    def get_atom_shell(self, atom_index: int) -> List[int]:
        # Get all shells for a specific atom in the PDOS data
        return list(set([orb['l'] for orb in self.projected_dos if orb['atom_index'] == atom_index]))
    
    def get_atom_shell_orbital(self, atom_index: int, l: int) -> List[str]:
        # Get all orbitals for a specific atom and shell in the PDOS data
        return list(set([orb['m'] for orb in self.projected_dos if (orb['atom_index'] == atom_index and orb['l'] == l)]))
    
    def plot_species_pdos(self, emin: float=-20, emax: float=10, pdos_fig_name: str='PDOS.png') -> None:
        #Plot PDOS for all species in the PDOS data.
        species = self.get_species()
        species_pdosdata = [self.get_pdos_by_species(species[i]) for i in range(len(species))]

        plot_dos_pdos([species_pdosdata],
                      [species],
                      titles=['Projected density of States of different species'],
                      energy=self.energy,
                      energy_min=emin,
                      energy_max=emax,
                      shifted=self.efermi is not None,
                      pdos_fig_name=pdos_fig_name)
    
    def plot_species_shell_pdos(self, emin: float=-20, emax: float=10, pdos_fig_name: str='PDOS.png') -> None:
        species = self.get_species()
        pdosdatas, labels, titles = [], [], []
        for species_i in species:
            species_shells = self.get_species_shell(species_i)
            pdosdatas.append([self.get_pdos_by_species_shell(species_i, species_shells[i]) for i in range(len(species_shells))])
            labels.append([f'{species_i}-{l_map[l]}' for l in species_shells])
            titles.append(f"PDOS for {species_i}")
        
        plot_dos_pdos(pdosdatas,
                      labels,
                      titles,
                      self.energy,
                      emin,
                      emax,
                      self.efermi is not None,
                      pdos_fig_name)
    
    def plot_species_orbital_pdos(self, emin: float=-20, emax: float=10, pdos_fig_name: str='PDOS.png') -> None:
        # Plot PDOS for all orbitals of all species in the PDOS data.
        species = self.get_species()
        pdosdatas, labels, titles = [], [], []
        for species_i in species:
            species_shells = self.get_species_shell(species_i)
            for shell in species_shells:
                species_shell_orbitals = self.get_species_shell_orbital(species_i, shell)
                pdosdatas.append([self.get_pdos_by_species_orbital(species_i, shell, species_shell_orbitals[i]) for i in range(len(species_shell_orbitals))])
                labels.append([f'{species_i}-{orbital_names[(shell, orbital)]}' for orbital in species_shell_orbitals])
                titles.append(f"PDOS for {species_i}-{l_map[shell]}")
        
        plot_dos_pdos(pdosdatas,
                      labels,
                      titles,
                      self.energy,
                      emin,
                      emax,
                      self.efermi is not None,
                      pdos_fig_name)
    
    def plot_atoms_pdos(self, atom_indices: List[int], emin: float=-20, emax: float=10, pdos_fig_name: str='PDOS.png') -> None:
        # Plot PDOS for all atoms specified in atom_indices.
        pdosdatas, labels, titles = [], [], []
        for i in atom_indices:
            species = self.get_atom_species(i)
            atom_shells = self.get_atom_shell(i)
            for shell in atom_shells:
                atom_shell_orbitals = self.get_atom_shell_orbital(i, shell)
                pdosdatas.append([self.get_pdos_by_atom_orbital(i, shell, orbital) for orbital in atom_shell_orbitals])
                labels.append([f'{species}{i}-{orbital_names[(shell, orbital)]}' for orbital in atom_shell_orbitals])
                titles.append(f"PDOS for {species}{i}-{l_map[shell]}")
        
        plot_dos_pdos(pdosdatas,
                      labels,
                      titles,
                      self.energy,
                      emin,
                      emax,
                      self.efermi is not None,
                      pdos_fig_name)

        
    def write_species_pdos(self, pdos_dat_file: str='PDOS.dat') -> None:
        #Write PDOS data of different species to a file.
        species = self.get_species()
        species_pdosdata = [self.get_pdos_by_species(species[i]) for i in range(len(species))]

        write_dos_pdos(species_pdosdata,
                       self.energy,
                       species,
                       self.efermi is not None,
                       pdos_dat_file)
    
    def write_species_shell_pdos(self, pdos_dat_file: str='PDOS.dat') -> None:
        species = self.get_species()
        pdosdatas, labels = [], []
        for species_i in species:
            species_shells = self.get_species_shell(species_i)
            pdosdatas.extend([self.get_pdos_by_species_shell(species_i, species_shells[i]) for i in range(len(species_shells))])
            labels.extend([f'{species_i}-{l_map[l]}' for l in species_shells])
        
        write_dos_pdos(pdosdatas,
                       self.energy,
                       labels,
                       self.efermi is not None,
                       pdos_dat_file)
    
    def write_species_orbital_pdos(self, pdos_dat_file: str='PDOS.dat') -> None:
        # Write PDOS data of different orbitals of different species to a file.
        species = self.get_species()
        pdosdatas, labels = [], []
        for species_i in species:
            species_shells = self.get_species_shell(species_i)
            for shell in species_shells:
                species_shell_orbitals = self.get_species_shell_orbital(species_i, shell)
                pdosdatas.extend([self.get_pdos_by_species_orbital(species_i, shell, species_shell_orbitals[i]) for i in range(len(species_shell_orbitals))])
                labels.extend([f'{species_i}-{orbital_names[(shell, orbital)].replace("$", "")}' for orbital in species_shell_orbitals])
        
        write_dos_pdos(pdosdatas,
                       self.energy,
                       labels,
                       self.efermi is not None,
                       pdos_dat_file)
    
    def write_atoms_pdos(self, atom_indices: List[int], pdos_dat_file: str='PDOS.dat') -> None:
        # Write PDOS data of different atoms to a file.
        pdosdatas, labels = [], []
        for i in atom_indices:
            species = self.get_atom_species(i)
            atom_shells = self.get_atom_shell(i)
            for shell in atom_shells:
                atom_shell_orbitals = self.get_atom_shell_orbital(i, shell)
                pdosdatas.extend([self.get_pdos_by_atom_orbital(i, shell, orbital) for orbital in atom_shell_orbitals])
                labels.extend([f'{species}{i}-{orbital_names[(shell, orbital)].replace("$", "")}' for orbital in atom_shell_orbitals])
        
        write_dos_pdos(pdosdatas,
                       self.energy,
                       labels,
                       self.efermi is not None,
                       pdos_dat_file)

    @staticmethod
    def sum_pdos_data(pdos_datas: List[Dict]) -> np.ndarray:
        """
        Sum the PDOS data from a list of extracted PDOS data.
    
        Parameters:
        -----------
        pdos_datas : list of dict
            List of PDOS data dictionaries
    
        Returns:
        --------
            Summed PDOS will be an array with shape (nspin, len(energy))
        """
        if len(pdos_datas) == 0:
            raise ValueError("No PDOS data provided")
    
        pdos_sum = np.zeros_like(pdos_datas[0]['data'])
        for pdos_data in pdos_datas:
            pdos_sum += np.array(pdos_data['data'])
        return pdos_sum


def plot_dos_pdos(pdosdatas: List[List[np.ndarray]],
                  labels: List[List[str]],
                  titles: List[str],
                  energy: np.ndarray,
                  energy_min: float,
                  energy_max: float,
                  shifted: bool,
                  pdos_fig_name: str):
    """
    Plot PDOS using given PDOS data, titles and labels.
    Args:
        - pdosdata (List[List[np.ndarray]]): PDOS datas used in the plot. The length of PDOS data is the number of subplots in the plot, and
          the length of each element of the list is number of PDOS data plotted in each subplot. If nspin=2, the spin up and spin down PDOS
          will be plotted in a subplot separately.
        - labels (List[List[str]]): Labels for each PDOS data in each subplot. The length of labels is the number of subplots in the plot, and
          the length of each element of the list is number of PDOS data plotted in each subplot. If nspin=2, up and down arrows will be added automatically.
        - titles (List[str]): Titles for each subplot. The length of titles is the number of subplots in the plot.
        - energy (np.ndarray): Energy values for the PDOS data.
        - energy_min (float): Minimum energy value for the plot.
        - energy_max (float): Maximum energy value for the plot.
        - pdos_fig_name (str): Name of the PDOS figure.
    """
    import matplotlib.pyplot as plt
    num_subplots = len(pdosdatas)

    START_MUTL_COL_SUBPLOT_NUM = 4
    TWO_COL_SUBPLOT_MAX_NUM = 6

    # Create subplots
    if num_subplots >= START_MUTL_COL_SUBPLOT_NUM:
        if num_subplots > TWO_COL_SUBPLOT_MAX_NUM:
            ncol = max(3, int(np.sqrt(num_subplots)))
        else:
            ncol = 2
        
        if num_subplots % ncol == 0:
            nrow = int(num_subplots/ncol)
        else:
            nrow = int(num_subplots/ncol) + 1
    else:
        nrow = num_subplots
        ncol = 1
    
    fig, axes = plt.subplots(nrow, ncol, figsize=(8*ncol, 4*nrow))
    if nrow * ncol == 1:
        axes = [axes]
    else:
        axes = axes.flatten()
    
    # Hide unused subplots to avoid displaying empty axes
    for idx in range(len(pdosdatas), len(axes)):
        axes[idx].set_visible(False)
    
    for idx, pdosdata in enumerate(pdosdatas):
        assert len(pdosdata) == len(labels[idx]) # Check if number of PDOS data matches number of labels
        ax = axes[idx]

        if_spin_polarized = False
        for i, (data, label) in enumerate(zip(pdosdata, labels[idx])):
            if data.shape[1] == 1: # nspin = 1 or 4
                ax.plot(energy, data, label=label, linestyle='-', color=color_list[i], linewidth=1.0)
            elif data.shape[1] == 2: # nspin = 2
                if_spin_polarized = True
                ax.plot(energy, data[:,0], label=f"{label} "+r"$\uparrow$", linestyle='-', color=color_list[i], linewidth=1.0)
                ax.plot(energy, -data[:,1], label=f"{label} "+r"$\downarrow$", linestyle='--', color=color_list[i], linewidth=1.0)
        
        ax.set_xlim(energy_min, energy_max)
        # calculate range of y axis for visible data
        y_min, y_max = [], []
        # find indices of energy in [energy_min, energy_max]
        visible_mask = (energy >= energy_min) & (energy <= energy_max)
        for data in pdosdata:
            if data.shape[1] == 1:
                visible_data = data[visible_mask]
                y_min.append(visible_data.min())
                y_max.append(visible_data.max())
            elif data.shape[1] == 2:
                visible_data_up = data[visible_mask, 0]
                visible_data_down = -data[visible_mask, 1]
                y_min.append(min(visible_data_up.min(), visible_data_down.min()))
                y_max.append(max(visible_data_up.max(), visible_data_down.max()))
        
        y_min = min(y_min)
        y_max = max(y_max)
        y_margin = (y_max - y_min) * 0.05  # Add 5% extra margin to the y axis
        if pdosdata[0].shape[1] == 1: # nspin = 1 or 4
            ax.set_ylim(0, y_max + y_margin)
        else: # nspin = 2
            ax.set_ylim(y_min - y_margin, y_max + y_margin)
        if shifted:
            ax.set_xlabel(r"$E-E_F$ (eV)", fontsize=12)
            ax.axvline(x=0, color="k", linestyle=":", alpha=0.5)
        else:
            ax.set_xlabel("Energy (eV)", fontsize=12)
        ax.set_ylabel("States", fontsize=12)
        ax.legend(loc='best', fontsize=8, ncol=if_spin_polarized+1)
        ax.grid(alpha=0.3)
        ax.set_title(titles[idx])
    
    plt.tight_layout()
    plt.savefig(pdos_fig_name, dpi=300)
    plt.close()

def write_dos_pdos(pdosdatas: List[np.ndarray], energy: np.ndarray, labels: List[str], shifted: bool, filename: str):
    """
    Write processed DOS and PDOS data to a file.

    Args:
        - pdosdata (List[np.ndarray]): Processed PDOS data. The shape of each element is same with self.projected_dos[i]['data'].
        - energy (np.ndarray): Energy values for the PDOS data.
        - labels (List[str]): Labels for each PDOS data. If nspin=2. up and down PDOS will be written in different columns.
        - filename (str): Name of the file to write the data to.
    """
    assert len(pdosdatas) == len(labels)
    
    min_width = 12

    energy_header = 'E-E_F(eV)' if shifted else 'Energy(eV)'
    column_headers = []
    for pdosdata, label in zip(pdosdatas, labels):
        if pdosdata.shape[1] == 1:
            column_headers.append(label)
        elif pdosdata.shape[1] == 2:
            column_headers.extend([label+'_up', label+'_dn'])
    
    # Calculate column widths
    col_widths = []
    energy_col_width = max(min_width, len(energy_header))
    col_widths.append(energy_col_width)
    for header in column_headers:
        calculated_width = max(min_width, len(header)+1)
        col_widths.append(calculated_width)

    with open(filename, "w") as writefile:
        header_line = f"{energy_header:>{col_widths[0]}s}"
        header_idx = 1
        for pdosdata, label in zip(pdosdatas, labels):
            if pdosdata.shape[1] == 1:
                header_line += f"{label:>{col_widths[header_idx]}s}"
                header_idx += 1
            elif pdosdata.shape[1] == 2:
                header_line += f"{label+'_up':>{col_widths[header_idx]}s}{label+'_dn':>{col_widths[header_idx+1]}s}"
                header_idx += 2
        writefile.write(header_line + "\n")

        for i in range(len(energy)):
            line = f"{energy[i]:>{col_widths[0]}.6f}"
            data_idx = 0
            for pdosdata in pdosdatas:
                if pdosdata.shape[1] == 1:
                    line += f"{pdosdata[i,0]:>{col_widths[data_idx+1]}.6f}"
                    data_idx += 1
                elif pdosdata.shape[1] == 2:
                    line += f"{pdosdata[i,0]:>{col_widths[data_idx+1]}.6f}{pdosdata[i,1]:>{col_widths[data_idx+2]}.6f}"
                    data_idx += 2
            writefile.write(line + "\n")



from .model import Model
import json, os
from abacustest.lib_prepare.abacus import WriteKpt, WriteInput, ReadInput, AbacusStru, ReadKpt
from abacustest.constant import RECOMMAND_IMAGE, RECOMMAND_COMMAND, RECOMMAND_MACHINE

from typing import List, Dict, Any

import shutil, glob, copy
from abacustest.lib_collectdata.collectdata import RESULT
from abacustest.lib_model.comm import clean_none_list


class BECModel(Model):
    @staticmethod
    def prepare_args(parser):
        '''
        Add arguments for the prepare subcommand
        The arguments can not be command, model, modelcommand 
        '''
        parser.description = "Prepare the inputs for Born effective charge calculation."
        parser.add_argument('-j', '--job', default=[], action="extend", nargs="*", help='the paths of ABACUS jobs, should contain INPUT, STRU, or KPT, and pseudopotential and orbital files')
        parser.add_argument("--stepsize", default=0.01, type=float, help="The step size for finite difference calculation, default is 0.01 Angstrom, which should be a small positive value.")
        parser.add_argument("--index", default=[0], nargs="*", type=int, help="The indices of atoms to calculate the Born effective charge, default is 0 (the first atom). You can specify multiple indices separated by space.")
        parser.add_argument("--dir", default=["x", "y", "z"], nargs="*", choices=["x", "y", "z"],type=str, help="The directions to displace the atoms, can be x, y, or z. Default is x y z. You can specify multiple directions separated by space.")
        parser.add_argument("--type", type=str, default="f", choices=["f", "b", "c"], help="The type of displacement, f: forward displacement, b: backward displacement, c: central difference (both forward and backward displacements). Default is f.")
        #parser.add_argument("--no-k-continuity", action="store_true", help="Whether to not set use_k_continuity in nscf calculations. Default is set. This parameter is used for ABACUS >= 3.10.1")
        parser.add_argument("--image", type=str, default=RECOMMAND_IMAGE, help="The image to use for the Bohrium job, default is %s" % RECOMMAND_IMAGE)
        parser.add_argument("--machine", type=str, default=RECOMMAND_MACHINE, help="The machine to use for the Bohrium job, default is 'c32_m64_cpu'.")
        parser.add_argument("--abacus_command", type=str, default=RECOMMAND_COMMAND, help=f"The command to run the Abacus job, default is '{RECOMMAND_COMMAND}'.")
        return parser
    
    
    def run_prepare(self,params):
        '''
        Run the model with the given parameters
        '''
        if not params.job:
            raise ValueError("No job specified, please use -j or --job to specify the job paths.")
        
        if params.stepsize <= 0:
            raise ValueError("Step size should be a positive value.")

        folders = prepare_bec_jobs(params.job,
                                   params.stepsize,
                                   params.index,
                                   params.dir,
                                   params.type,
                                   k_continuity=False)
        # write run scripts
        with open("run.sh", "w") as f1:
            f1.write(f"cp INPUT.scf INPUT\ncp KPT.scf KPT\n{params.abacus_command} | tee scf.log\n\n")
            for i in range(1,4):
                f1.write(f"cp INPUT.nscf{i} INPUT\ncp KPT.nscf{i} KPT\n{params.abacus_command} | tee nscf{i}.log\n")
                f1.write(f"mv OUT.ABACUS/running_nscf.log OUT.ABACUS/running_nscf{i}.log\nrm OUT.ABACUS/*.restart\n")
        
        setting = {
            "save_path": "results",
            "run_dft": [{
                "ifrun": True,
                "example": folders,
                "extra_files":["run.sh"],
                "command": "bash run.sh",
                "image": params.image,
                "bohrium": {
                    "scass_type": params.machine,
                    "job_type": "container",
                    "platform": "ali"
                }
            }]
        } 
        
        setting_file = "setting.json"
        json.dump(setting, open(setting_file, "w"), indent=4)
        print("\nThe inputs are generated in", ", ".join(params.job))
        print(f"You can modify '{setting_file}' and 'run.sh', and execute below command to run the abacustest to submit all jobs to bohrium:\n\tabacustest submit -p setting.json\n")
        print(f"After finishing the calculations, you can enter the results directory, \nand run below command below to postprocess the BEC results:\n\tabacustest model bec post -j {' '.join(params.job)}\n")
        
    
    def postprocess_args(parser):
        '''
        Add arguments for the postprocess subcommand
        The arguments can not be command, model, modelcommand 
        '''
        parser.description = "Postprocess the elastic calculation results."
        parser.add_argument('-j', '--job', default=[], action="extend", nargs="*", help='the paths of the job directories, should contain the results of deformed structures generated by the prepare step.')
        return parser   

    def run_postprocess(self,params):
        '''
        Parse the parameters and run the postprocess process'''
        if not params.job:
            raise ValueError("No job specified, please use -j or --job to specify the job paths.")
        
        metrics, becs = postprocess_bec(params.job)
        json.dump(becs, open("metrics_bec.json", "w"), indent=4)
        json.dump(metrics, open("metrics.json", "w"), indent=4)
        print("\nThe BEC results are saved in 'metrics_bec.json'")

        import pandas as pd
        c = "Born effective charge tensor for:"
        for k, v in becs.items():
            c += f"\n{k}:\n"
            bect, index = clean_none_list(v["bec_tensor (e)"], ["X", "Y", "Z"])
            pd_df = pd.DataFrame(bect, columns=['X', 'Y', 'Z'], index=index)
            # save 4 decimal places
            pd_df = pd_df.round(4)
            c += f"{pd_df}\n"

        c += """
NOTE: The BEC tensor is in unit of e, calculated by delta P / delta R, 
where P is polarization vector in cartesian coordinates (e/Volume * A), 
and R is atomic position in cartesian coordinates (A). 
Displacement is the actual atomic displacement in cartesian coordinates (A).
The p_vec_disp and p_vec_org are polarization vectors along the three cell vectors.
The p1(C/m^2) is the polarization in C/m^2 along the three cell vectors, and p2(C/m^2) is the polarization in C/m^2 in the cartesian coordinates.
        """
        with open("bec_summary.txt", "w") as f1:
            f1.write(c)
        print("\nSummary of Born effective charge tensors:")
        print(c)
            
        
        
def prepare_bec_jobs(jobs: str,
                     stepsize: float,
                     index: List[int],
                     directions: List[str],
                     disp_type: str = "f",
                     k_continuity: bool = True
                    ) -> None:
    '''
    Prepare the BEC calculation jobs.
    
    Args:
        jobs: List of job paths.
        stepsize: Step size for finite difference.
        index: List of atom indices to displace.
        directions: List of directions to displace ('x', 'y', 'z').
        disp_type: Type of displacement, 'f' for forward, 'b' for backward, 'c' for central difference.
    '''
    if disp_type not in ["f", "b", "c"]:
        raise ValueError(f"Invalid displacement type: {disp_type}. Should be 'f', 'b', or 'c'.")

    if stepsize == 0:
        raise ValueError("Step size should be a non-zero value.")
    if stepsize < 0:
        stepsize = -stepsize
    
    directions = list(set(directions))
    directions_map = {"x": [stepsize, 0.0, 0.0],
                      "y": [0.0, stepsize, 0.0],
                      "z": [0.0, 0.0, stepsize]}
    if set(directions) - set(directions_map.keys()):
        raise ValueError(f"Invalid directions specified: {directions}. Valid options are 'x', 'y', 'z'.")

    folders = []
    for job in jobs:
        # clean org folders
        print("Preparing BEC calculation for job path:", job)
        old_folders = [f for f in glob.glob(os.path.join(job, "bec_*")) if os.path.isdir(f)]
        if len(old_folders) > 0:
            print(f"Remove old bec_* folders in job path: {job}")
        for f in old_folders:
            shutil.rmtree(f)
        
        if not os.path.isfile(os.path.join(job, "STRU")):
            raise FileNotFoundError(f"STRU file not found in job path: {job}")
        if not os.path.isfile(os.path.join(job, "INPUT")):
            raise FileNotFoundError(f"INPUT file not found in job path: {job}")
        
        org_files = [f for f in os.listdir(job) if f not in ["INPUT", "STRU", "KPT"]]
        input_param = ReadInput(os.path.join(job, "INPUT"))
        stru = AbacusStru.ReadStru(os.path.join(job, "STRU"))
        kpt, kpt_model = ReadKpt(job)
        input_param.pop("kspacing", None) # we should remove kspacing, but use KPT instead
        input_param["suffix"] = "ABACUS"
        # if SPIN1_CHG.cube exists, use it as initial charge density
        if os.path.isfile(os.path.join(job, "OUT.ABACUS/SPIN1_CHG.cube")):
            input_param["init_chg"] = "file"
        
        if disp_type != "c":
            org_path = os.path.join(job, f"bec_org")
            os.makedirs(org_path, exist_ok=True)
            for f in org_files:
                os.symlink(os.path.abspath(os.path.join(job, f)), os.path.join(org_path, f))
            write_inputs(org_path, stru, input_param, kpt, kpt_model, k_continuity=k_continuity)
            folders.append(org_path)
        

        for direction in directions:           
            # write displaced structure folders
            for idx in index:
                disp_vec = directions_map[direction]

                if disp_type in ["f", "c"]:
                    disp_stru = copy.deepcopy(stru)
                    atom_coord = disp_stru.get_coord(bohr=False, direct=False)
                    atom_coord[idx] = [atom_coord[idx][i] + disp_vec[i] for i in range(3)]
                    disp_stru.set_coord(atom_coord, bohr=False, direct=False)
                    disp_path = os.path.join(job, f"bec_disp_atom{idx}_{direction}")
                    os.makedirs(disp_path, exist_ok=True)
                    for f in org_files:
                        os.symlink(os.path.abspath(os.path.join(job, f)), os.path.join(disp_path, f))
                    write_inputs(disp_path, disp_stru, input_param, kpt, kpt_model, k_continuity=k_continuity)
                    folders.append(disp_path)
                
                if disp_type in ["b", "c"]:
                    disp_stru = copy.deepcopy(stru)
                    atom_coord = disp_stru.get_coord(bohr=False, direct=False)
                    atom_coord[idx] = [atom_coord[idx][i] - disp_vec[i] for i in range(3)]
                    disp_stru.set_coord(atom_coord, bohr=False, direct=False)
                    disp_path = os.path.join(job, f"bec_disp_atom{idx}_{direction}_back")
                    os.makedirs(disp_path, exist_ok=True)
                    for f in org_files:
                        os.symlink(os.path.abspath(os.path.join(job, f)), os.path.join(disp_path, f))
                    write_inputs(disp_path, disp_stru, input_param, kpt, kpt_model, k_continuity=k_continuity)
                    folders.append(disp_path)

    return folders
            
            
def write_inputs(folder, stru, input_param, kpt, kpt_model, k_continuity: bool = True):
    # 1. write INPUT
    input_param_copy = copy.deepcopy(input_param)
    input_param_copy["calculation"] = "scf"
    input_param_copy["out_chg"] = 1
    input_param_copy["out_bandgap"] = 1
    WriteInput(input_param_copy, os.path.join(folder, "INPUT.scf"))
    
    # 2. write KPT
    WriteKpt(kpt, os.path.join(folder, "KPT.scf"),kpt_model)
    
    # 3. write INPUT nscf
    input_param_copy["calculation"] = "nscf"
    input_param_copy["out_chg"] = None
    input_param_copy["init_chg"] = "file"
    input_param_copy["berry_phase"] = 1
    if k_continuity:
        input_param_copy["use_k_continuity"] = True
    for i in range(1,4):
        input_param_copy["gdir"] = i
        WriteInput(input_param_copy, os.path.join(folder, f"INPUT.nscf{i}"))
    
    # 4. write KPT nscf
    for i in range(3):
        kpt_nscf = copy.deepcopy(kpt)
        kpt_nscf[i] = 2 * kpt[i] # double the k-point grid in the displacement direction
        WriteKpt(kpt_nscf, os.path.join(folder, f"KPT.nscf{i+1}"),kpt_model)
    
    # 5. write STRU
    stru.write(os.path.join(folder, "STRU"))

    
def postprocess_bec(jobs: List[str]) -> Dict[str, Any]:
    '''
    Postprocess the BEC calculation results.
    
    Args:
        jobs: List of job paths.
    
    Returns:
        A dictionary containing the BEC results.
    '''
    xyz_idx = {"x":0, "y":1, "z":2}
    metrics = {}
    becs = {}
    for job in jobs:
        print("Postprocessing BEC calculation for job path:", job)
        if not os.path.isdir(job):
            print(f"Warning: job path {job} is not a directory, skip.")
            continue

        sub_folders = [f for f in glob.glob(os.path.join(job, "bec_*")) if os.path.isdir(f)]
        sub_folders.sort()
        label = AbacusStru.ReadStru(os.path.join(sub_folders[0], "STRU")).get_label(total=True)
        
        for folder in sub_folders:
            metrics[folder] = read_metrics(folder)
            stru = AbacusStru.ReadStru(os.path.join(folder, "STRU"))
            atom_coord = stru.get_coord(bohr=False, direct=False)
            cell = stru.get_cell(bohr=False)
            p_vec1, mod1, volume1, p_cm2_1 = read_p(os.path.join(folder, "OUT.ABACUS/running_nscf1.log"))
            p_vec2, mod2, volume2, p_cm2_2 = read_p(os.path.join(folder, "OUT.ABACUS/running_nscf2.log"))
            p_vec3, mod3, volume3, p_cm2_3 = read_p(os.path.join(folder, "OUT.ABACUS/running_nscf3.log"))
            # check volume consistency
                                           
            metrics[folder]["atom_coord"] = atom_coord
            metrics[folder]["cell"] = cell
            metrics[folder]["p_vec"] =  [p_vec1, p_vec2, p_vec3]  # p_vec1/2/3 is along cell vectors
            metrics[folder]["p1(C/m^2)"] = [p_cm2_1, p_cm2_2, p_cm2_3] # along a/b/c directions
            metrics[folder]["p2(C/m^2)"] = cal_polarization_cartesian([p_cm2_1, p_cm2_2, p_cm2_3], cell)  # in cartesian coordinates
            metrics[folder]["volume"] = volume1
            metrics[folder]["mod"] = [mod1, mod2, mod3]
            
        
        for folder in sub_folders:
            # bec_disp_atom{idx}_{direction}
            if "bec_disp" not in os.path.basename(folder):
                continue
            parts = os.path.basename(folder).split("_")
            idx = int(parts[2][4:])
            direction = parts[3]

            org_folder = os.path.join(job, f"bec_org")
            # find the org_folder
            if folder.endswith("_back") and folder[:-5] in sub_folders:
                continue  # will be handled in the positive displacement
            elif (not folder.endswith("_back")) and (folder + "_back" in sub_folders):
                org_folder = folder + "_back"

            if org_folder not in metrics:
                print(f"Warning: folder {org_folder} not found for displaced folder {folder}. Skipping.")
                continue
            
            p_vec_disp = metrics[folder]["p_vec"]
            p_vec_org = metrics[org_folder]["p_vec"]
            if p_vec_disp is None or p_vec_org is None:
                print(f"Warning: Polarization vector not found for folder {folder} or {org_folder}. Skipping.")
                continue
            
            coord_disp = metrics[folder]["atom_coord"][idx]
            coord_org = metrics[org_folder]["atom_coord"][idx]
            delta_r_atom = [(coord_disp[i] - coord_org[i]) for i in range(3)]
            disp_idx = xyz_idx[direction]
            delta_r_norm = sum([delta_r_atom[i]**2 for i in range(3)])**0.5 * (1 if delta_r_atom[disp_idx] >=0 else -1)
            delta_p = cal_delta_p(p_vec_org, p_vec_disp, metrics[org_folder]["mod"], metrics[folder]["mod"])
            delta_p = cal_polarization_cartesian(delta_p, metrics[org_folder]["cell"])  # transform to cartesian
            bec_tensor = [delta_p[i] / delta_r_norm for i in range(3)]
            
            ilabel = f"{job}:atom{idx}_{label[idx]}"

            if ilabel not in becs:
                becs[ilabel] = {
                    "bec_tensor (e)": [None, None, None],
                    "displacement (A)": [None, None, None],
                    "p_vec_org ((e/Volume)*A)": p_vec_org,
                    "mod_org ((e/Volume)*A)": metrics[org_folder]["mod"],
                    "p_vec_disp ((e/Volume)*A)": [None, None, None]
                }

            becs[ilabel]["bec_tensor (e)"][xyz_idx[direction]] = bec_tensor
            becs[ilabel]["displacement (A)"][xyz_idx[direction]] = delta_r_atom
            becs[ilabel]["p_vec_disp ((e/Volume)*A)"][xyz_idx[direction]] = p_vec_disp
    
    return metrics, becs

def cal_polarization_cartesian(p_vec, cell):
    '''
    Convert polarization vector along cell vectors to cartesian coordinates.
    
    Args:
        p_vec: Polarization vector in cell vector basis.
        cell: Cell vectors.
    '''
    import numpy as np
    cell_matrix = np.array(cell)
    p_cartesian = np.sum([p_vec[i] * cell_matrix[i] / np.linalg.norm(cell_matrix[i]) for i in range(3)], axis=0)

    return p_cartesian.tolist()

def cal_delta_p(p1, p2, mod1, mod2):
    '''
    Calculate the delta p considering the polarization quantum.
    
    Args:
        p1: Polarization vector 1.
        p2: Polarization vector 2.
        mod1: Modulus for polarization vector 1.
        mod2: Modulus for polarization vector 2.
    '''
    delta_p = [p2[i] - p1[i] for i in range(3)]
    for i in range(3):
        if delta_p[i] > 0:
            if abs(delta_p[i] - mod2[i]) < abs(delta_p[i]):
                delta_p[i] = delta_p[i] - mod2[i]
        else:
            if abs(delta_p[i] + mod2[i]) < abs(delta_p[i]):
                delta_p[i] = delta_p[i] + mod2[i]
    return delta_p

def read_metrics(job: str) -> List[str]:
    '''
    Read the available metrics from the job results.
    
    Args:
        job: Path to the job directory.
        
    Returns:
        A list of available metric names.
    '''
    shutil.copy(os.path.join(job, "INPUT.scf"), os.path.join(job, "INPUT"))
    r = RESULT(path=job, fmt="abacus",output=os.path.join(job, "scf.log"))
    keys = ["normal_end", "converge","energy", "scf_steps", "denergy_last", "drho_last", "band_gap"]
    return {k: r[k] for k in keys}

def read_p(logf):
    """Read polarization from log file.
    
    Args:
        logf: Path to the log file.
        
    Returns:
        p_vec: List of polarization vector components (P * V, with unit in e*A)
    
    """
    """
        The calculated polarization direction is in R1 direction

 P =    4.6793707  (mod    4.7318739)  (   4.6793707,   0.0000000,   0.0000000) (e/Omega).bohr

 P =    0.0212834  (mod    0.0215222)  (   0.0212834,   0.0000000,   0.0000000) e/bohr^2

 P =    1.2168069  (mod    1.2304597)  (   1.2168069,   0.0000000,   0.0000000) C/m^2
    """
    from abacustest.constant import BOHR2A
    with open(logf) as f1: lines = f1.readlines()  
    
    p_vec = None
    volume = None
    mod = None
    p_cm2 = None
    for i, line in enumerate(lines):
        if "The calculated polarization direction is " in line:
            p_vec = float(lines[i+2].split()[2]) * BOHR2A
            mod = float(lines[i+2].split("mod")[1].split(")")[0]) * BOHR2A
            p_cm2 = float(lines[i+6].split()[2])  # in C/m^2
        elif "Volume (A^3) =" in line:
            volume = float(line.split("=")[1])
    return p_vec, mod, volume, p_cm2        
                    



import os,sys,glob,json,shutil,copy,traceback
from abacustest import constant
from abacustest.lib_prepare.abacus import AbacusStru,ReadInput,WriteInput,WriteKpt
from abacustest.lib_collectdata.collectdata import RESULT
from abacustest.prepare import PrepareAbacus
import numpy as np
from .model import Model
from . import comm,comm_magj
from abacustest.constant import RECOMMAND_IMAGE, RECOMMAND_COMMAND, RECOMMAND_MACHINE

PREPARE_DOC="""
    This model is used to prepare the calculation of the magnetic exchange interactions (J) of two atom.
    You should prepare a file named magj.txt in the folder, which define the atoms to calculate the J.
    Each line of the file defines a pair of atoms with format: <Atomlabel1> <AtomIndex1> <Atomlabel2> <AtomIndex2>
    The AtomIndex is start from 1 in each type of atom.  
"""

POST_DOC="""
    This model is used to postprocess the calculation of the magnetic exchange interactions (J) of two atom.
    You can get the J from the output files.
"""

class magj(Model):
    '''
    Base class for all models
    
    Some specific workflow will be implemented in the subclass.
    Such as EOS, BAND, etc.
    
    The model should have the following methods:
    '''
    @staticmethod
    def add_args(parser):
        '''
        Add arguments for the model
        The arguments can not be command, model, modelcommand 
        '''
        pass
    
    def run(self,params):
        '''
        Parse the parameters and run the model
        '''
        pass
    
    @staticmethod
    def prepare_args(parser):
        '''
        Add arguments for the prepare subcommand
        The arguments can not be command, model, modelcommand '''
        
        parser.description = PREPARE_DOC
        parser.add_argument('-d', '--step', type=float, default=1,help="the tilt step. Default 1 degree.")
        parser.add_argument('-n', '--number', type=int, default=5,help="the tilt times. Default is 5 times. The maximum tilt degree is step*number")
        parser.add_argument('-j', '--job',default=["."], action="extend",nargs="*" ,help='the path of abacus inputs, default is current folder')
        parser.add_argument('-f', '--file', type=str, default="magj.txt",help="the file name of atom info, which define which two atoms to calculate the J. Default is magj.txt")
        
        return parser
    
    def run_prepare(self,params):
        '''
        Parse the parameters and run the prepare process.
        Usually, this step will generate the input files for abacustest submit.
        '''
        jobs = ["."] if len(params.job) == 1 else params.job[1:]
        infof = "magj.txt"
        subfolders = PrepareMagJ(infof,jobs,params.step,params.number).run()

        if subfolders:
            setting = {
                "save_path": "results",
                "bohrium_group_name": "MagJ",
                "run_dft": {
                    "example": subfolders,
                    "command": RECOMMAND_COMMAND,
                    "image": RECOMMAND_IMAGE,
                    "bohrium": {
                        "scass_type": RECOMMAND_MACHINE,
                        "job_type": "container",
                        "platform": "ali",
                    },
                },
            }
            comm.dump_setting(setting)
            comm.doc_after_prepare("MagJ", subfolders, ["setting.json"],has_prepare=False)
    
    @staticmethod
    def postprocess_args(parser):
        '''
        Add arguments for the postprocess subcommand
        The arguments can not be command, model, modelcommand'''
        
        parser.description = POST_DOC
        parser.add_argument('-j', '--job',default=["."], action="extend",nargs="*" ,help='the path of abacus inputs, default is current folder.')
        return parser

    def run_postprocess(self,params):
        '''
        Parse the parameters and run the postprocess process'''
        jobs = ["."] if len(params.job) == 1 else params.job[1:]
        PostProcessMagJ(jobs,"magj").run()


def Condition2folder(subfolder, label1, label1idx, label2, label2idx, angle, case):   
    # CASE0: magj_Fe1_Fe2_angle1_1
    # CASE1: magj_Fe1_Fe2_angle1_2
    # CASE2: magj_Fe1_Fe2_angle1_3
    # 1: tilt atom1, 2: tilt atom2, 3: tilt atom1 and atom2 together
    # check: abacustest/lib_model/comm_magj.py
    angle = str(angle).rstrip("0").rstrip(".")
    return f"{subfolder}_{label1}{label1idx+1}_{label2}{label2idx+1}_angle{angle}_{case+1}"  

def Folder2Condition(folder):
    def split_label(label):
        if label[1].isdigit():
            return label[0],int(label[1])-1
        else:
            return label[:2],int(label[2:])-1
    try:
        ks = folder.split("_")
        subfolder = ks[0]
        label1, idx1 = split_label(ks[1])
        label2, idx2 = split_label(ks[2])
        angle = float(ks[3][5:])
        case = int(ks[4])-1
        return (subfolder, label1, idx1, label2, idx2, angle, case)
    except:
        traceback.print_exc()
        print("ERROR: folder name format error in",folder)
        return None

    
class PrepareMagJ:
    def __init__(self,infof,jobs,step,number):
        self.infof = infof
        self.jobs = jobs
        self.step = step
        self.number = number
        self.subfolder = "magj"
    
    def run(self):
        step = self.step
        num = self.number
        print("step (degree):",self.step)
        print("number of steps:",self.number)

        subfolders = []
        for job in self.jobs:
            for ijob in glob.glob(job):
                if not os.path.isdir(ijob):
                    continue
                # remove magj folders
                for ifile in glob.glob(os.path.join(ijob,self.subfolder+"*")):
                    if os.path.isdir(ifile):
                        shutil.rmtree(ifile)
                print("prepare inputs in ",ijob)
                infos = self.read_info(ijob)
                if infos:
                    isub = self.preapre_inputs(infos,ijob,step,num=num)
                    if isub:
                        subfolders.extend(isub)
                else:
                    print("ERROR: preapre failed in",ijob)
        # write relative path of subfolders to example.txt
        return subfolders
    
    def read_info(self,path):
        """
        the txt.info should be like:
        Fe 1 Fe 2
        
        Return:
        [["Fe", 0, "Fe", 1]]
        """
        def print_error(path,line):
            print(f"ERROR: {self.infof} format error in ",path)
            print("line:",line)
            print(PREPARE_DOC)
        if os.path.isfile(os.path.join(path,self.infof)):
            with open(os.path.join(path,self.infof)) as f:
                lines = f.readlines()
                info = []
                for line in lines:
                    if line.strip() == "":
                        continue
                    sline = line.strip().split()
                    if len(sline) == 4:
                        info.append([sline[0],int(sline[1])-1,sline[2],int(sline[3])-1])
                        if info[-1][1] < 0 or info[-1][3] < 0:
                            print_error(path,line)
                            return None
                    else:
                        print_error(path,line)
                        return None
                return info
        else:
            print(f"ERROR: {self.infof} not found in ",path)
            return None
        
    def prepare_one_stru_abacus(self,istru,final_path,otherfiles,input_param):
        os.makedirs(final_path,exist_ok=True)
        istru.write(os.path.join(final_path,"STRU"))
        # copy other files to final_path
        for file in otherfiles:
            os.symlink(file,os.path.join(final_path,os.path.basename(file)))

        # write input
        PrepareAbacus.WriteInput(input_param,os.path.join(final_path,"INPUT"))

    def preapre_inputs(self,infos, init_path, step = 1.0, num=3) :
        subfolders = []
        input_param = {"calculation": "scf"}
        otherfiles = []
        for ifile in os.listdir(init_path):
            if ifile == "INPUT":
                input_param = PrepareAbacus.ReadInput(os.path.join(init_path,"INPUT"))
                input_param["calculation"] = "scf"
                continue
            elif ifile != "STRU" and not ifile.startswith(".") and os.path.isfile(os.path.join(init_path,ifile)):
                otherfiles.append(os.path.abspath(os.path.join(init_path,ifile)))

        stru = AbacusStru.ReadStru(os.path.join(init_path,"STRU"))
        if not stru:
            print(f"ERROR: read STRU failed in {init_path}")
            return None

        label = stru.get_label(total=True)
        atom_mag = stru.get_atommag()
        new_stru = copy.deepcopy(stru)

        # write original input
        self.prepare_one_stru_abacus(stru,os.path.join(init_path,self.subfolder),otherfiles,input_param)
        #with open(os.path.join(init_path,subfolder,"step.txt"),"w") as f: f.write(str(step))
        subfolders.append(os.path.join(init_path,self.subfolder))
        
        for info in infos:
            label1, index1, label2, index2 = info
            if label1 not in label or label2 not in label:
                print(f"ERROR: label {label1} or {label2} not found in STRU in {init_path}")
                continue
            global_idx1 = 0 
            global_idx2 = 0
            for i in range(len(label)):
                if label[i] == label1 and label[:i+1].count(label1) == index1+1:
                    global_idx1 = i
                if label[i] == label2 and label[:i+1].count(label2) == index2+1:
                    global_idx2 = i
            mag1 = atom_mag[global_idx1]
            mag2 = atom_mag[global_idx2]
            mags = comm_magj.prepare_atom_mag(mag1, mag2, step, num)
            for i, mag in enumerate(mags):
                angle = (i+1)*step
                
                for j in range(3):
                    atom_mag[global_idx1] = mag[j][0]
                    atom_mag[global_idx2] = mag[j][1]
                    new_stru.set_atommag(atom_mag)
                    subpath = os.path.join(init_path, Condition2folder(self.subfolder, label1, index1, label2, index2, angle,j ))
                    self.prepare_one_stru_abacus(new_stru,
                                                 subpath,
                                                 otherfiles,input_param)
                    subfolders.append(subpath)

        return subfolders
    
class PostProcessMagJ:
    def __init__(self,job,subfolder="magj"):
        self.job = job
        self.subfolder = "magj"
    
    def run(self):
        all_metrics = {}
        all_results = {}
        for ijob in self.job:
            if not os.path.isdir(ijob):
                continue
            print("postprocess in ",ijob)
            metrics = self.get_metrics(ijob)
            json.dump(metrics,open(os.path.join(ijob,"metrics.json"),"w"),indent=4)
            all_metrics.update(metrics)
            
            results = self.metrics2results(metrics)
            json.dump(results,open(os.path.join(ijob,"results.json"),"w"),indent=4)
            all_results[ijob] = results
        
        json.dump(all_metrics,open("metrics.json","w"),indent=4)
        json.dump(all_results,open("results.json","w"),indent=4)
        # print the J
        for k,v in all_results.items():
            for kk,vv in v.items():
                if kk == "original_energy (eV)":
                    continue
                print(f"{k} {kk} J:",vv["j (meV)"])
                
    
    def get_metrics(self, path):
        # find the subfolders in path
        # the subfolder name is like "magj_He1_He2_angle1_1"
        metrics = {}
        for ipath in glob.glob(os.path.join(path, self.subfolder+"*")):
            if not os.path.isdir(ipath):
                continue
            basename = os.path.basename(ipath)
            if not (basename == self.subfolder or len(basename.split("_")) == 5):
                continue
            
            r = RESULT(path=ipath, fmt="abacus")
            metrics[ipath] = {
                "converge": r["converge"],
                "energy": r["energy"],
                "drho_last": r["drho_last"],
                "denergy_last": r["denergy_last"],
                "denergy_womix_last": r["denergy_womix_last"],
                "atomlabel_list": r["atomlabel_list"],
                "total_time": r["total_time"] ,
                "mag": r["ds_mag"]
            }
        return metrics
    
    def get_atom_mag(self, atomlabel_list, mag, atomlabel, atomindex):
        """
        Get the magnetic moment of the atoms in atomlabel_list
        
        Parameters
        ----------
        atomlabel_list : list of str
            The list of atom label of all atoms
        mag : list of lists
            The magnetic moment of all atoms, which should respect the order of atomlabel_list
        atomlabel : str
            The atom label to get the magnetic moment
        atomindex : int
            The atom index of this atom type
        """
        idx = 0
        for i, label in enumerate(atomlabel_list):
            if label == atomlabel:
                if idx == atomindex:
                    return mag[i]
                idx += 1
        return None
    
    def metrics2results(self, metrics):
        """
        Convert the metrics to results
        
        Parameters  
        ----------
        metrics : dict
            The metrics of the calculation. The key is the subfolder name, and the value is the metrics of the calculation.
            
        Returns
        -------
        results : dict
            The results that is convient to calculate the J
        
        results = {
            "original_energy": -100,  # the energy of the original structure
            "Fe1_Fe2":{
                "angle": [1,2,3,4,5], # the target tilt angle
                "angle_cos": [[0.1,0.2,0.3],...] # the cos of the tilt angle, which is the average of the angle_cos_detail
                "angle_detail": [[1,1,1],...] # the detail of the tilt angle of atom1, atom2, and atom1 and atom2 together
                "angle_cos_detail": [[0.1,0.2,0.3],...] # the cos of the tilt angle of atom1, atom2, and atom1 and atom2 together
                "energy": [[-100,-101,-102],...] # the energy of tilt angle of atom1, atom2, and atom1 and atom2 together
                "delta_energy": [1,2,3,4,5], # (E_togather - E_original) - (E_atom1 - E_original) - (E_atom2 - E_original)
            }
        """
        metrics_new = {os.path.basename(k):v for k,v in metrics.items()}
        
        results = {
            "original_energy (eV)": metrics_new.get(self.subfolder,{}).get("energy",None)
        }
        for k,v in metrics_new.items():
            if k == self.subfolder:
                continue
            
            conditions = Folder2Condition(k)
            if conditions is None:
                continue
            
            _, label1, idx1, label2, idx2, tilt_angle, job_idx = conditions
            # job_idx: 0: tilt atom1, 1: tilt atom2, 2: tilt atom1 and atom2 together
            
            case_name = f"{label1}{idx1+1}_{label2}{idx2+1}"
            if case_name not in results:
                results[case_name] = {
                    "angle": [],
                    "angle_detail": [],
                    "angle_cos_detail": [],
                    "energy (eV)": [],
                }
            if tilt_angle not in results[case_name]["angle"]:
                results[case_name]["angle"].append(tilt_angle)
                results[case_name]["angle_detail"].append([None,None,None])
                results[case_name]["angle_cos_detail"].append([None,None,None])
                results[case_name]["energy (eV)"].append([None,None,None])
                
            angle_idx = results[case_name]["angle"].index(tilt_angle)
            mag1 = self.get_atom_mag(v["atomlabel_list"], v["mag"], label1, idx1)
            mag2 = self.get_atom_mag(v["atomlabel_list"], v["mag"], label2, idx2)
            if None not in [mag1,mag2]:
                results[case_name]["angle_detail"][angle_idx][job_idx] = np.arccos(np.dot(mag1,mag2)/np.linalg.norm(mag1)/np.linalg.norm(mag2))*180/np.pi
                results[case_name]["angle_cos_detail"][angle_idx][job_idx] = np.dot(mag1,mag2)/np.linalg.norm(mag1)/np.linalg.norm(mag2)
                
            results[case_name]["energy (eV)"][angle_idx][job_idx] = v["energy"]
        
        e_original = results["original_energy (eV)"]
        for k,v in results.items():
            if k == "original_energy (eV)":
                continue
            results[k]["angle_cos"] = []
            results[k]["delta_energy (eV)"] = []
            for i in range(len(v["angle"])):
                # calcualte the averge of the angle_cos_detail
                angle_cos_detail = v["angle_cos_detail"][i]
                if None in angle_cos_detail:
                    results[k]["angle_cos"].append(None)
                else:
                    results[k]["angle_cos"].append(np.mean(angle_cos_detail))
                
                # calculate the delta_energy    
                e1 = v["energy (eV)"][i][0]
                e2 = v["energy (eV)"][i][1]
                etot = v["energy (eV)"][i][2]
                if None in [e1,e2,etot,e_original]:
                    results[k]["delta_energy (eV)"].append(None)
                else:
                    results[k]["delta_energy (eV)"].append((etot - e_original) - (e1 - e_original) - (e2 - e_original))
                    
            results[k]["j (meV)"] = self.cal_j(results[k]["delta_energy (eV)"], results[k]["angle_cos"])
        return results
    
    def cal_j(self, deltae, cos_theta):
        """
        Calculate the J from the results.
        Will fit: \delta E = J * (1 - cos(\theta))
        
        Parameters
        ----------
        results : dict
            The results that is convient to calculate the J
        
        Returns
        -------
        j : dict
            The J of the atoms
        """
        
        # clean the None value
        de = []
        ct = []
        for i in range(len(deltae)):
            if deltae[i] is not None and cos_theta[i] is not None:
                de.append(deltae[i])
                ct.append(1 - cos_theta[i])
        
        if len(de) < 1:
            return None
        else:
            # fit the J by polyfit
            j = np.polyfit(ct,de,1)[0]
            return j * 1000
            
            
            

        
        


from .model import Model
from . import comm
import argparse,json, os
from abacustest.lib_prepare.abacus import WriteKpt, WriteInput, gen_stru, ReadInput, AbacusStru
from pathlib import Path
from abacustest.lib_model.model_012_band import PrepBand
from abacustest.constant import RECOMMAND_IMAGE, RECOMMAND_COMMAND, RECOMMAND_MACHINE

import warnings

from typing import List, Dict, Union, Optional, Tuple, Literal


JOB_TYPES = {"scf": {"calculation": "scf", "symmetry": 1, "ecutwfc": 80, "scf_thr": 1e-8, "scf_nmax": 100,
                    "smearing_method": "gauss", "smearing_sigma": 0.015, "mixing_type": "broyden",
                    "mixing_beta": 0.8,  "basis_type": "pw  # or lcao", "ks_solver": "dav_subspace  # or genelpa for lcao basis",
                    "pw_diag_ndim":    2, "pw_diag_nmax":    20,
                    "precision": "double  # or single",
                    "#cal_force": 1, "#cal_stress": 1,
                    "kspacing": "0.14 # unit in 1/bohr"}, 
             "relax": {"calculation": "relax", "symmetry": 1, "ecutwfc": 80, "scf_thr": 1e-8, "scf_nmax": 100,
                       "smearing_method": "gauss", "smearing_sigma": 0.015, "mixing_type": "broyden",
                       "mixing_beta": 0.8, "basis_type": "pw  # or lcao", "ks_solver": "dav_subspace  # or genelpa for lcao basis",
                       "pw_diag_ndim":    2, "pw_diag_nmax":    20,
                       "precision": "double  # or single",
                       "cal_force": 1, "#cal_stress": 1,"kspacing": "0.14 # unit in 1/bohr",
                       "relax_method": "cg # or bfgs bfgs_trad cg_bfgs sd fire",
                       "relax_nmax": 60, "force_thr_ev": "0.01  # unit in eV/A", "#stress_thr": "0.5 # unit in kbar",
                       }, 
             "cell-relax":{"calculation": "cell-relax", "symmetry": 1, "ecutwfc": 80, "scf_thr": 1e-8, "scf_nmax": 100,
                       "smearing_method": "gauss", "smearing_sigma": 0.015, "mixing_type": "broyden",
                       "mixing_beta": 0.8, "basis_type": "pw  # or lcao", "ks_solver": "dav_subspace  # or genelpa for lcao basis",
                       "pw_diag_ndim":    2, "pw_diag_nmax":    20,
                       "precision": "double  # or single",
                       "cal_force": 1, "cal_stress": 1, "kspacing": "0.14 # unit in 1/bohr",
                       "relax_method": "cg # or bfgs, bfgs_trad, cg_bfgs, sd, fire",
                       "relax_nmax": 60, "force_thr_ev": "0.01  # unit in eV/A", "stress_thr": "0.5 # unit in kbar",
                       "fixed_axes": "None # or volume, shape, a, b, c, ab, ac, bc; only valid for cell-relax calculation to fix some axes",
                       }, 
             "md":{"calculation": "md", "symmetry": 1, "ecutwfc": 80, "scf_thr": 1e-8, "scf_nmax": 100,
                       "smearing_method": "gauss", "smearing_sigma": 0.015, "mixing_type": "broyden",
                       "mixing_beta": 0.8, "basis_type": "pw  # or lcao", "ks_solver": "dav_subspace  # or genelpa for lcao basis",
                       "pw_diag_ndim":    2, "pw_diag_nmax":    20,
                       "precision": "double  # or single",
                       "#cal_force": 1, "#cal_stress": 1,
                       "kspacing": "0.14 # unit in 1/bohr", 
                       "md_type": "nvt  # or npt, nve, langevin, fire, msst",
                       "md_nstep": "10  # number of steps",
                       "md_dt": "1.0  # unit in fs", 
                       "md_tfirst": "100  # unit in K",
                       "md_tlast": "100  # unit in K",}, 
             "band":{"calculation": "scf", "symmetry": 1, "ecutwfc": 80, "scf_thr": 1e-8, "scf_nmax": 100,
                    "smearing_method": "gauss", "smearing_sigma": 0.015, "mixing_type": "broyden",
                    "mixing_beta": 0.8,  "basis_type": "pw  # or lcao", "ks_solver": "dav_subspace  # or genelpa for lcao basis",
                    "pw_diag_ndim":    2, "pw_diag_nmax":    20,
                    "precision": "double  # or single",
                    "#cal_force": 1, "#cal_stress": 1,
                    "kspacing": "0.14 # unit in 1/bohr"}
        }

LCAO_PARAM = {
    "basis_type": "lcao",
    "ks_solver": "genelpa",
    "#gamma_only": 0,
    "ecutwfc": 100,
    "scf_thr": 1e-7,
}

ELEMENT_DFTU_D = ["Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn",
                  "Y", "Zr", "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd",
                  "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", "La", "Ac", "Th"]
ELEMENT_DFTU_F = ["Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu",
                  "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr"]

MAG_ELEMENT = ELEMENT_DFTU_D + ELEMENT_DFTU_F


class InputsModel(Model):
    HAS_PREPARE_POST_COMMAND = False # set false if the model does not have prepare and postprocess

    @staticmethod
    def add_args(parser):
        '''
        Add arguments for the prepare subcommand
        The arguments can not be command, model, modelcommand '''
        
        parser.description = "Prepare the ABACUS inputs file."
        parser.add_argument('-f', '--file',default=[], action="extend",nargs="*" ,help='the structure files')
        parser.add_argument("--ftype",default="cif",type=str,help="the structure type, should be cif or dpdata supportted type like: poscar, abacus/stru, ..", )
        parser.add_argument("--jtype", default="scf", type=str, help=f"the job type, should be one of {list(JOB_TYPES.keys())}")
        parser.add_argument("--pp",default=None,type=str,help="the path of pseudopotential library, or read from enviroment variable ABACUS_PP_PATH")
        parser.add_argument("--orb",default=None,type=str,help="the path of orbital library, or read from enviroment variable ABACUS_ORB_PATH")
        parser.add_argument("--input",default=None,type=str,help="the template of input file, if not specified, the default input will be generated")
        parser.add_argument("--kpt", default=None, type=int, nargs="*", help="the kpoint setting, should be one or three integers")
        parser.add_argument("--lcao", action="store_true", help="whether to use lcao basis, default is pw basis")
        parser.add_argument("--nspin", default=1, type=int, choices=[1, 2, 4], help="the number of spins, can be 1 (no spin), 2 (spin polarized), or 4 (non-collinear spin). Default is 1.")
        parser.add_argument("--soc", action="store_true", help="whether to use spin-orbit coupling, if True, nspin should be 4.")
        parser.add_argument("--dftu", action="store_true", help="whether to use DFT+U, default is False.")
        parser.add_argument("--dftu_param", default=None, nargs="+", help="the DFT+U parameters, should be element symbol and U value pairs like 'Fe 4 Ti 1'. If dftu is set, but dftu_param is not set, the default U values will be used: 4 eV for d orbital elements and 6 eV for f orbital elements.")
        parser.add_argument("--init_mag", default=None, nargs="+", help="the initial magnetic moment for magnetic elements, should be element symbol and magnetic moment pairs like 'Fe 4 Ti 1'.")
        parser.add_argument("--afm", action="store_true", help="whether to use antiferromagnetic calculation, default is False. Only valid when init_mag is set.")
        parser.add_argument("--copy_pp_orb", action="store_true", help="whether to copy the pseudopotential and orbital files to each job directory or link them. Default is False, which means linking the files.")
        parser.add_argument("--folder-syntax", default=None, type=str, help="A python syntax to create the directory name. Using string variable 'x' to represent the path of input structure file, and put the processing of x in {}, following the format of f-string. Such as: '{x[:-4]}' to remove the file extension like Fe.cif to Fe, and 'aa-{x[:-4]}-bb' will transfer Fe.cif to aa-Fe-bb. Default is None, which means using number like 000000/000001...")
        parser.add_argument("--overwrite", action="store_true", help="Whether to overwrite existing folder (default: False, create new folder with numbered suffix if exists)")
        parser.add_argument("--download-pporb", nargs="?", type=str,default=None,const="apns-v1",choices=["apns-v1"], help="Download the recommended pseudopotential and orbital files from AISquare.")
        return parser
    
    @staticmethod
    def parse_dftu_param(values):
        """
        Parse the values from the command line arguments.
        """
        if len(values) % 2 != 0:
            raise ValueError("The DFT+U parameters should be in pairs of element symbol and U value.")
        dftu_param = {}
        for i in range(0, len(values), 2):
            element = values[i]
            u_value = float(values[i + 1])
            if element not in dftu_param:
                dftu_param[element] = u_value
            else:
                warnings.warn(f"Element {element} already has a DFT+U value, overwriting it with {u_value}.")
        return dftu_param
     
    @staticmethod
    def parse_init_mag(values):
        """
        Parse the initial magnetic moment values from the command line arguments.
        """
        if len(values) % 2 != 0:
            raise ValueError("The initial magnetic moment parameters should be in pairs of element symbol and magnetic moment.")
        init_mag = {}
        for i in range(0, len(values), 2):
            element = values[i]
            mag_value = float(values[i + 1])
            if element not in init_mag:
                init_mag[element] = mag_value
            else:
                warnings.warn(f"Element {element} already has an initial magnetic moment, overwriting it with {mag_value}.")
        return init_mag   

    def run(self,params):
        if params.download_pporb is not None:
            download_abacus_pporb(params.download_pporb)
            return 0
        
        if params.kpt is not None and len(params.kpt) not in [1, 3]:
            raise ValueError("The kpoint setting should be one or three integers.")
        
        if params.dftu_param is not None:
            dftu_param = self.parse_dftu_param(params.dftu_param)
        else:
            dftu_param = None
            
        if params.init_mag is not None:
            init_mag = self.parse_init_mag(params.init_mag)
        else:
            init_mag = None
        
        if not params.file:
            raise ValueError("Please specify the structure file.")
        
        pinput = PrepInput(
            files=params.file,
            filetype=params.ftype,
            jobtype=params.jtype,
            pp_path=params.pp,
            orb_path=params.orb,
            input_file=params.input,
            kpt=params.kpt,
            lcao=params.lcao,
            nspin=params.nspin,
            soc=params.soc,
            dftu=params.dftu,
            dftu_param=dftu_param,
            init_mag=init_mag,
            afm=params.afm,
            abacus_command=RECOMMAND_COMMAND,
            machine=RECOMMAND_MACHINE,  # default Bohrium machine type for CPU jobs
            image=RECOMMAND_IMAGE,  # default recommended image for ABACUS jobs 
            copy_pp_orb=params.copy_pp_orb, 
            folder_syntax=params.folder_syntax,
            overwrite=params.overwrite
        )
        pinput.run()
        return 0

def download_abacus_pporb(version):
    """
    Download the recommended pseudopotential and orbital files from AISquare.
    """
    import shutil

    if version is None:
        return
    if version == "apns-v1":
        url = "https://store.aissquare.com/datasets/af21b5d9-19e6-462f-ada1-532f47f165f2/ABACUS-APNS-PPORBs-v1.zip"
        target_name = "ABACUS-APNS-PPORBs-v1.zip"
    else:
        raise ValueError(f"Invalid version: {version}")
    download_file = comm.download_file(url, target_name)
    out_files = None
    if download_file is not None:
        print("Downloaded file:", download_file)
        out_files = comm.smart_extract(download_file, ".")

        if out_files is not None:
            os.remove(download_file)
            print("Extracted files:", out_files)
            print("You can set the pseudopotential and orbital path by --pp and --orb, or set the environment variable ABACUS_PP_PATH and ABACUS_ORB_PATH.")
    
    return out_files

class PrepInput:
    """Prepare the ABACUS inputs for the specified model.
    This class will generate the ABACUS inputs files based on the specified structure files, job type, and other parameters.
    
    Parameters
    ----------
    files : list
        The structure files, should be a list of cif or dpdata supported data
    filetype : str
        The structure type, cif or dpdata supported format
    pp_path : str
        The pseudopotential path, if not specified, will read from the environment variable ABACUS_PP_PATH
    orb_path : str
        The orbital path, if not specified, will read from the environment variable ABACUS_ORB_PATH
    input_file : str
        The template of input file, if not specified, the default input will be generated
    kpt : list
        The kpoint setting, should be a list of three int, will generate a KPT file
    abacus_command : str
        The command to run ABACUS, default is RECOMMAND_COMMAND
    machine : str
        The machine type, default is RECOMMAND_MACHINE, which is the Bohrium machine type for CPU jobs.
    image : str
        The Docker image to use, default is RECOMMAND_IMAGE, which is the recommended image for ABACUS jobs.
    lcao : bool
        Whether to use LCAO basis, default is False, which means using PW basis.
    nspin : int
        The number of spins, can be 1 (no spin), 2 (spin polarized), or 4 (non-collinear spin). Default is 1.
    soc : bool
        Whether to use spin-orbit coupling, if True, nspin should be 4.
    dftu : bool
        Whether to use DFT+U, default is False.
    dftu_param : dict
        The DFT+U parameters, should be a dict like {"Fe": 4, "Ti": 1}, where the key is the element symbol and the value is the U value.
        Value can also be a list of two values, and the first value is the orbital (p, d, f) to apply DFT+U, and the second value is the U value.
        For example, {"Fe": ["d", 4], "O": ["p", 1]} means applying DFT+U to Fe 3d orbital with U=4 eV and O 2p orbital with U=1 eV.
        If dftu is True, but dftu_param is None, the default U values will be used: 4 eV for d orbital elements and 6 eV for f orbital elements.
    init_mag : dict or None
        the initial magnetic moment for magnetic elements, should be a dict like {"Fe": 4, "Ti": 1}, where the key is the element symbol and the value is the initial magnetic moment.
    afm : bool
        Whether to use antiferromagnetic calculation, default is False. If True, half of the magnetic elements will be set to negative initial magnetic moment.
    copy_pp_orb : bool
        Whether to copy the pseudopotential and orbital files to each job directory or link them. Default is False, which means linking the files.
    folder_syntax: str, the python syntax for each structure corresponding to its folder, 
                   using string variable "x" to denote the input structure filename, and write the python syntax in {}. 
                   For example, if the structure file is "Fe.cif", the syntax "{x[:-4]}" represents "Fe", and "aa-{x[:-4]}-yy" represents "aa-Fe-yy". 
                   This function will create the new directory and save the structure inside it. 
                   The default is None, indicating the generation of 00001, 00002
                   If the folder already exists, the function will add a number to the folder name. 
                   Like Fe.1, Fe.2, ...
    overwrite: bool, Whether to overwrite existing folder (default: False, create new folder with numbered suffix if exists)
    """
    
    def __init__(self, files: Union[str, List[str], Path, List[Path]],
                 filetype: str, 
                 jobtype: str, 
                 pp_path: Optional[Union[str, Path]]=None, 
                 orb_path: Optional[Union[str, Path]]=None, 
                 input_file: Optional[Union[str, Path]]=None, 
                 kpt: Optional[Tuple[int, int, int]]=None,
                 abacus_command: str=RECOMMAND_COMMAND, 
                 machine: str=RECOMMAND_MACHINE, 
                 image: str=RECOMMAND_IMAGE,
                 lcao: bool=False,
                 nspin: Literal[1,2,4]=1,  # can be 1 or 2 or 4, 1: no spin, 2: spin polarized, 4: non-collinear spin
                 soc: bool =False, # spin-orbit coupling, if True, nspin should be 4 
                 dftu: bool=False,
                 dftu_param: Optional[Dict[str, Union[float, Tuple[Literal["s", 'p', "d"], float]]]]=None,
                 init_mag: Optional[Dict[str, float]] =None,
                 afm: bool = False, 
                 copy_pp_orb: bool = False,
                 folder_syntax: Optional[str] = None,
                 overwrite: bool = False
                 ):
        if jobtype not in JOB_TYPES:
            raise ValueError(f"Unsupported job type: {jobtype}.\nSupported job types are {list(JOB_TYPES.keys())}.")
        
        if not isinstance(files, list):
            self.files = [files]
        else:
            self.files = files
        
        # check the parameters
        if nspin not in [1, 2, 4]:
            raise ValueError(f"Invalid nspin value: {nspin}. It should be 1 (no spin), 2 (spin polarized), or 4 (non-collinear spin).")
        if soc and nspin != 4:
            warnings.warn("Spin-orbit coupling is set, but nspin is not 4. Setting nspin to 4 for non-collinear spin calculation.")
            nspin = 4

        self.filetype = filetype
        self.jobtype = jobtype
        self.input_file = input_file
        self.kpt = kpt
        self.abacus_command = abacus_command
        self.machine = machine
        self.image = image
        self.lcao = lcao
        self.nspin = nspin
        self.soc = soc
        self.dftu = dftu
        self.dftu_param = dftu_param
        self.init_mag = init_mag
        self.afm = afm
        self.copy_pp_orb = copy_pp_orb
        self.folder_syntax = folder_syntax
        self.overwrite = overwrite
        
        self.pp_path = pp_path
        self.orb_path = orb_path
        if pp_path is None and os.environ.get("ABACUS_PP_PATH") is not None:
            self.pp_path = os.environ["ABACUS_PP_PATH"]
        if orb_path is None and os.environ.get("ABACUS_ORB_PATH") is not None:
            self.orb_path = os.environ["ABACUS_ORB_PATH"]

        print("Structure files:", self.files)
        print("Structure type:", self.filetype)
        print("Job type:", self.jobtype)
        print("Pseudopotential path:", self.pp_path)
        print("Orbital path:", self.orb_path)
        print("Input file:", self.input_file)
        print("Kpoint:", self.kpt)
        print("LCAO basis:", self.lcao)
        print("Nspin:", self.nspin)
        print("SOC:", self.soc)
        print("DFTU:", self.dftu)
        print("DFTU parameters:", self.dftu_param)
        print("Initial magnetic moment:", self.init_mag)
        print("AFM:", self.afm)
        print("ABACUS command:", self.abacus_command)
        print("Machine:", self.machine)
        print("Image:", self.image)
        print("")
    
    def run(self):
        jobs = gen_stru(self.files, self.filetype, self.pp_path, self.orb_path, tpath=".", 
                        copy_pp_orb=self.copy_pp_orb,
                        folder_syntax=self.folder_syntax,
                        overwrite=self.overwrite)

        if self.input_file is not None:
            if not os.path.isfile(self.input_file):
                raise ValueError(f"Input file {self.input_file} not found.")
            else:
                input_template = ReadInput(self.input_file)
        else:
            input_template = None
        
        for job in jobs:
            recommand_ecutwfc = jobs[job].get("recommand_ecutwfc", None)
            element = jobs[job]["element"]
            
            input_param = self.generate_input(element)
            input_param = self.update_input(input_param, input_template, recommand_ecutwfc, element)
            
            WriteInput(input_param, os.path.join(job, "INPUT"))
            self.set_init_mag(job, input_param.get("nspin", 1))
            self.write_kpt(job)
        
        job_path = list(jobs.keys())
        if self.jobtype == "band":
            prep_band = PrepBand(job_path, run_command=self.abacus_command)
            _, run_script, extra_files = prep_band.run()
        else:
            run_script_file = "run.sh"
            Path(run_script_file).write_text(self.abacus_command)
            run_script = f"bash {run_script_file}"
            extra_files = [run_script_file]
            

        # generate the parameter setting for abacustest submit
        setting = {
            "save_path": "results",
            "run_dft": {
                "example": job_path,
                "extra_files": extra_files,
                "command": run_script,
                "image": self.image,
                "bohrium": {
                    "scass_type": self.machine,
                    "job_type": "container",
                    "platform": "ali"
                }
            }
        }
        
        setting_file = "setting.json"
        json.dump(setting, open(setting_file, "w"), indent=4)
        print("\nThe inputs are generated in", ", ".join(job_path))
        print(f"Please modify the ABACUS running command in script {extra_files[0]} if needed.")
        print(f"You can modify '{setting_file}', and execute the command 'abacustest submit -p setting.json' to run the abacustest to submit all jobs to bohrium.")
        print(f"Or you can 'cd' to each job path and execute '{extra_files[0]}' to run the job.")
        
        return setting, job_path

    
    def write_kpt(self, job):
        if self.kpt is not None:
            WriteKpt(self.kpt, os.path.join(job, "KPT"))
    
    def set_init_mag(self, job, nspin):
        """
        Set the initial magnetic moment for the job.
        If afm is True, half of the magnetic elements will be set to negative initial magnetic moment.
        """
        if not self.init_mag and not self.afm:
            return
        
        noncolin = nspin == 4
        
        stru_file = os.path.join(job, "STRU")
        if not os.path.isfile(stru_file):
            raise ValueError(f"Structure file {stru_file} not found.")
        stru = AbacusStru.ReadStru(stru_file)
        element = stru.get_element(number=False, total=True)
        
        if self.afm and not self.init_mag:
            mag_element = list(set([e for e in element if e in MAG_ELEMENT]))
            if len(mag_element) == 0:
                raise ValueError(f"No magnetic elements found in the structure {stru_file}, but AFM is set to True. Please set init_mag or remove afm option.")
            else:
                print(f"Setting initial magnetic moment (3 uB) for antiferromagnetic calculation for elements: {mag_element}.")
            init_mag = {e: 3 for e in mag_element}  # default initial magnetic moment is 1 for all magnetic elements
        else:
            init_mag = self.init_mag
        
        atom_mag = [0 for _ in element]
        atommag_idx = {}
        for i, e in enumerate(element):
            if e in init_mag:
                if e not in atommag_idx:
                    atommag_idx[e] = []
                atommag_idx[e].append(i)
                atom_mag[i] = init_mag[e]
        
        if len(atommag_idx) == 0:
            print(f"No magnetic elements found in the structure {stru_file}, skipping initial magnetic moment setting.")
            return
        
        if self.afm:
            # set half of the magnetic elements to negative initial magnetic moment
            print("Setting initial magnetic moment for antiferromagnetic calculation.")
            for e in atommag_idx:
                idx = atommag_idx[e]
                if len(idx) > 1:
                    for i in idx[::2]:
                        atom_mag[i] = -atom_mag[i]
                    print(f"Odd indexed atoms of element {e} are set to negative magnetic moment.")

        if noncolin:
            atom_mag = [[0,0,i] for i in atom_mag]
            print(f"Non-collinear spin calculation, set initial magnetic moment with three components [0, 0, m] for each atom.")
        
        stru.set_atommag(atom_mag)
        stru.write(stru_file)
        
    
    def update_input(self, input_param, input_template, recommand_ecutwfc, element):
        """
        Update the input parameters based on the input template and recommended ecutwfc.
        """
        if input_template is not None: 
            if "calculation" in input_template:
                calculation = input_template["calculation"]
                if calculation != self.jobtype:
                    warnings.warn(f"The calculation type in the input template is {calculation}, but the job type is {self.jobtype}. The input template will be ignored.")
                input_template.pop("calculation")
            input_param.update(input_template)
        else:
            input_template = {}
        
        # only when input_template has not set ecutwfc, and the basis_type is pw, we will reset the ecutwfc based on the recommand_ecutwfc
        if "ecutwfc" not in input_template and input_param.get("basis_type").startswith("pw"):
            if recommand_ecutwfc is not None and list(set(recommand_ecutwfc)) != [None]:
                input_param["ecutwfc"] = max([i for i in recommand_ecutwfc if i is not None])
                print(f"Set ecutwfc to {input_param['ecutwfc']}, element: {element}, recommended ecutwfc: {recommand_ecutwfc}")
        
        if not input_param.get("basis_type", "pw").startswith("pw"):
            # remove pw specific parameters
            input_param.pop("pw_diag_ndim", None)
            input_param.pop("pw_diag_nmax", None)
            
        return input_param
    
    def generate_input(self, element):
        """
        Set the special input parameters for the job.
        """
        input_param = JOB_TYPES[self.jobtype].copy()
        pw_basis = input_param.get("basis_type", "pw").startswith("pw")
        
        # LCAO
        if self.lcao:
            input_param.update(LCAO_PARAM)
            input_param.pop("pw_diag_ndim", None)
            input_param.pop("pw_diag_nmax", None)
        
        # NSPIN
        if self.nspin == 2:
            input_param["nspin"] = 2
            input_param["mixing_beta"] = 0.4
            input_param["symmetry"] = 0
            input_param["onsite_radius"] = 3  # to generate the atomic magnetic moment
            input_param["out_mul"] = 1  # to output the mulliken charge
            
        if self.soc or self.nspin == 4:
            if self.nspin != 4:
                warnings.warn("nspin is set to 4 for non-collinear spin calculation.")
            input_param["nspin"] = 4
            input_param["noncolin"] = 1
            input_param["mixing_beta"] = 0.4
            input_param["symmetry"] = -1
            input_param["onsite_radius"] = 3
            input_param["out_mul"] = 1  # to output the mulliken charge
            if self.soc:
                input_param["lspinorb"] = 1
        
        # DFTU
        if self.dftu:
            if self.nspin == 1:
                warnings.warn("DFTU is set, but nspin is 1. We suggest to set nspin to 2 or 4 for DFTU calculation.")
            input_param["dft_plus_u"] = 1
            orbital_corr = []
            hubbard_u = []
            for e in element:
                # set u for element defined in dftu_param
                if self.dftu_param is not None and e in self.dftu_param:
                    if isinstance(self.dftu_param[e], (list,tuple)):
                        orbital = {"p":1, "d":2, "f":3}.get(self.dftu_param[e][0], None)  
                        if orbital is None:
                            raise ValueError(f"Unsupported orbital type {self.dftu_param[e][0]} for element {e}. Supported types are p, d, f.")
                        orbital_corr.append(orbital)
                        hubbard_u.append(self.dftu_param[e][1])
                    else:
                        if e in ELEMENT_DFTU_D:
                            orbital_corr.append(2)
                        elif e in ELEMENT_DFTU_F:
                            orbital_corr.append(3)
                        else:
                            orbital_corr.append(1)
                        hubbard_u.append(self.dftu_param[e])
                # if not define dftu_param, then set u for d and f orbital elements
                elif self.dftu_param is None:
                    if e in ELEMENT_DFTU_D:
                        orbital_corr.append(2)
                        hubbard_u.append(4)  # default U value for d orbital
                    elif e in ELEMENT_DFTU_F:
                        orbital_corr.append(3)
                        hubbard_u.append(6)  # default U value for f orbital
                    else:
                        orbital_corr.append(-1)
                        hubbard_u.append(0)  # no DFTU for other elements
                else:
                    orbital_corr.append(-1)
                    hubbard_u.append(0)  # no DFTU for other elements
            input_param["orbital_corr"] = orbital_corr
            input_param["hubbard_u"] = hubbard_u
            input_param["onsite_radius"] = 3
            if len(set(orbital_corr)) > 1 or orbital_corr[0] != -1:
                if pw_basis:
                    input_param["uramping"] = max(hubbard_u)
                else:
                    input_param["mixing_dmr"] = 1
            
            input_param["mixing_restart"] = 0.001
        
        if input_param.get("basis_type", "pw").startswith("pw") and input_param.get("out_mul", ""):
            # out_mul is only valid for lcao basis
            input_param.pop("out_mul")
            
        if input_param.get("basis_type", "pw").startswith("pw") and input_param.get("nspin",1) == 2 and "onsite_radius" in input_param:
            # onsite_radius is invalid for pw basis with nspin=2
            input_param.pop("onsite_radius")

        return input_param
    


import traceback
from dp.launching.typing.basic import BaseModel, Int, String, Float, List, Optional, Union, Dict
from dp.launching.typing import (
    BaseModel,
    Set,
    Boolean,
    Field,
    InputFilePath
)
from enum import Enum
from typing import Literal
import re,os,shutil
from . import comm_class_exampleSource,comm_func


class PPLibEnum(String, Enum):
    dataset1 = "SG15-ONCV-v1_0"

class OrbLibEnum(String, Enum):
    dataset1 = "SG15-Version1_0__StandardOrbitals-Version2_0"

class PPFromDatasets(BaseModel):
    type: Literal["from datasets"]
    dataset: PPLibEnum = Field(title="PP (pseudo potential) Lib datasets",
                        description="Please choose the PP Lib datasets.")
    '''
    dataset_unrecorded: String = Field(default=None,
                                       description="By default, the pp lib \"[SG15-ONCV-v1_0](https://launching.mlops.dp.tech/?request=GET%3A%2Fapplications%2Fabacustest%2Fdatasets%2Fpporb)\" will be used. \
If you want to use other pplib dataset, please enter download link here.")
    '''

class OrbFromDatasets(BaseModel):
    type: Literal["from datasets"]
    dataset: OrbLibEnum = Field(title="Orb Lib datasets",
                        description="Please choose the Orb Lib datasets.")
    '''
    dataset_unrecorded: String = Field(default=None,
                                       description="By default, the Orb lib \"[SG15-Version1_0__StandardOrbitals-Version2_0](https://launching.mlops.dp.tech/?request=GET%3A%2Fapplications%2Fabacustest%2Fdatasets%2Fpporb)\" will be used. \
If you want to use other pplib dataset, please enter download link here.")
    '''

#prepare INPUT template
class PrepareInputTemplateSet(BaseModel):
    PrepareInputTemplate_local: InputFilePath = Field(default=None,
                                                  title="Upload INPUT template locally",
                                                  description="""You can upload an INPUT file, and then all examples will use this INPUT""")

# prepare INPUT template path, for dataset case
class PrepareInputTemplatePathSet(BaseModel):
    PrepareInputTemplatePath: String = Field(default=None,title="INPUT template",description = "If you want to use an INPUT template for all examples, please enter the path of INPUT template here.")             

#prepare STRU template
class PrepareStruTemplateSet(BaseModel):
    PrepareStruTemplate_local: InputFilePath = Field(default=None,
                                                  title="Upload STRU template locally",
                                                  description="""You can upload a STRU file, and then all examples will use this STRU""")

#prepare STRU template path, for dataset case
class PrepareStruTemplatePathSet(BaseModel):
    PrepareStruTemplatePath: String = Field(default=None,title="STRU template",description = "If you want to use an STRU template for all examples, please enter the path of STRU template here.")
            
#prepare KPT template
class PrepareKptTemplateSet(BaseModel):
    PrepareKptTemplate_local: InputFilePath = Field(default=None,
                                                  title="Upload KPT template locally",
                                                  description="""You can upload a KPT file, and then all examples will use this KPT""")

#prepare KPT template path, for dataset case
class PrepareKptTemplatePathSet(BaseModel):
    PrepareKptTemplatePath: String = Field(default=None,title="KPT template",description = "If you want to use an KPT template for all examples, please enter the path of KPT template here.")

#prepare dpks descriptor file
class PrepareDPKSDescriptorSet(BaseModel):
    PrepareDPKSDescriptor_local: InputFilePath = Field(default=None,
                                                  title="Upload DeeP-KS Descriptor locally",
                                                  description="""If you want to use Deep-KS, you can upload the descriptor file here, or prepare the descriptor file in each example directory.""")

# prepare dpks descriptor path, for dataset case
class PrepareDPKSDescriptorPathSet(BaseModel):
    PrepareDPKSDescriptorPath: String = Field(default=None,title="DeeP-KS Descriptor",description = "If you want to use an DeeP-KS Descriptor for all examples, please enter the path of DeeP-KS Descriptor here.")
    
#prepare pp lib
class PreparePPLibSet(BaseModel):
    PreparePPLib_local: InputFilePath = Field(default=None,
                                                  title="Upload pseudopotential library locally",
                                                  description="""If you have a pseudopotential library and you want to use it for all elements, please upload it here. Please also prepare an \"element.json\" file in the library directory, and the key is name of element and value is the name of pseudopotential file.""")
    '''
    PreparePPLib: Union[
        comm_class_exampleSource.NotRquired,
        comm_class_exampleSource.FromPreUpload,
    #    comm_class_exampleSource.FromDatahub,
        PPFromDatasets] = Field(title="Prepare pseudopotential lib source",
                              discriminator="type",
                              description="Please choose the PP Lib source.")   
    '''
# prepare pp lib path, for dataset case
class PreparePPLibPathSet(BaseModel):
    PreparePPLibPath: String = Field(default=None,title="PP Lib",description = "If you want to use pseudopotential library, please enter the path of PP Lib here.")
    
#prepare ORB lib
class PrepareOrbLibSet(BaseModel):
    PrepareOrbLib_local: InputFilePath = Field(default=None,
                                                  title="Upload Orbital library locally",
                                                  description="""If you have an Orbital library, you can upload it here. Please also prepare a \"element.json\" file in the library directory, and the key is name of element and value is the name of Orbital file.""")
    '''
    PrepareOrbLib: Union[
        comm_class_exampleSource.NotRquired,
        comm_class_exampleSource.FromPreUpload,
    #    comm_class_exampleSource.FromDatahub,
        OrbFromDatasets] = Field(title="Prepare Orb Lib source",
                              discriminator="type",
                              description="Please choose the Orb Lib source.")    
    '''
# prepare orb lib path, for dataset case
class PrepareOrbLibPathSet(BaseModel):
    PrepareOrbLibPath: String = Field(default=None,title="Orb Lib",description = "If you want to use Orbital library, please enter the path of Orb Lib here.")

class PrepareSet(BaseModel):
    
    prepare_mix_kpt: String = Field(default=None,title="Additional KPT settings",description="You can set additional KPT SETTING for each examples. \
Please enter 1/3/6 values seperated by space (such as \"2\" means [2,2,2,0,0,0], which is 3 K ponits and the shift of mesh in K space; \"1 2 3\" means [1,2,3,0,0,0]; \"1 2 3 1 0 0\" means [1,2,3,1,0,0]).\
You can set multiple values (separated by comma), which will generate a set of ABACUS inputs for each value. (Scuch as: \"3, 1 2 3\")")

    prepare_mix_input: Optional[Dict[String,String]] = Field(default=None,title="Additional INPUT settings",description="You can set additional INPUT parameters for each examples. \
You can set multiple values (separated by comma) for each parameter, which will generate a set of ABACUS inputs for each value. If you want to combine several parameters, you can connect the parameters with |, and also conbine values with |. Such as: key is ecutwfc|kspacing, and value is 50|0.1, 60|0.2, 70|0.3. Commonly used parameters: \
\"calculation\", \"ecutwfc\", \"scf_thr\", \"scf_nmax\", \"basis_type\", \"smearing_method\", \"smearing_sigma\", \"mixing_type\", \"mixing_beta\",\"ks_solver\"")
    
def parse_prepare(prepare_set,work_path,download_path,logs):
    prepare = {}
    
    dataset_work_path = comm_class_exampleSource.get_dataset_work_path(prepare_set)
            
    # parse INPUT template
    if hasattr(prepare_set,"PrepareInputTemplate_local") and getattr(prepare_set,"PrepareInputTemplate_local") \
        and os.path.isfile(prepare_set.PrepareInputTemplate_local.get_path()):
            prepare["input_template"] = prepare_set.PrepareInputTemplate_local.get_path()
    elif dataset_work_path and hasattr(prepare_set,"PrepareInputTemplatePath") and getattr(prepare_set,"PrepareInputTemplatePath") and \
        os.path.isfile(os.path.join(dataset_work_path,prepare_set.PrepareInputTemplatePath)):
            prepare["input_template"] = os.path.join(dataset_work_path,prepare_set.PrepareInputTemplatePath)  
    
    # parse STRU template
    if hasattr(prepare_set,"PrepareStruTemplate_local") and getattr(prepare_set,"PrepareStruTemplate_local") and \
        os.path.isfile(prepare_set.PrepareStruTemplate_local.get_path()):
            prepare["stru_template"] = prepare_set.PrepareStruTemplate_local.get_path()
    elif dataset_work_path and hasattr(prepare_set,"PrepareStruTemplatePath") and getattr(prepare_set,"PrepareStruTemplatePath") and \
        os.path.isfile(os.path.join(dataset_work_path,prepare_set.PrepareStruTemplatePath)):
            prepare["stru_template"] = os.path.join(dataset_work_path,prepare_set.PrepareStruTemplatePath)
    
    # parse KPT template
    if hasattr(prepare_set,"PrepareKptTemplate_local") and getattr(prepare_set,"PrepareKptTemplate_local") and \
        os.path.isfile(prepare_set.PrepareKptTemplate_local.get_path()):
            prepare["kpt_template"] = prepare_set.PrepareKptTemplate_local.get_path()
    elif dataset_work_path and hasattr(prepare_set,"PrepareKptTemplatePath") and getattr(prepare_set,"PrepareKptTemplatePath") and \
        os.path.isfile(os.path.join(dataset_work_path,prepare_set.PrepareKptTemplatePath)):
            prepare["kpt_template"] = os.path.join(dataset_work_path,prepare_set.PrepareKptTemplatePath)
    
    # parse DPKS descriptor
    if hasattr(prepare_set,"PrepareDPKSDescriptor_local") and getattr(prepare_set,"PrepareDPKSDescriptor_local") and \
        os.path.isfile(prepare_set.PrepareDPKSDescriptor_local.get_path()):
            prepare["dpks_descriptor"] = prepare_set.PrepareDPKSDescriptor_local.get_path()
    elif dataset_work_path and hasattr(prepare_set,"PrepareDPKSDescriptorPath") and getattr(prepare_set,"PrepareDPKSDescriptorPath") and \
        os.path.isfile(os.path.join(dataset_work_path,prepare_set.PrepareDPKSDescriptorPath)):
            prepare["dpks_descriptor"] = os.path.join(dataset_work_path,prepare_set.PrepareDPKSDescriptorPath)
            
    # parse mix input
    if hasattr(prepare_set,"prepare_mix_input") and prepare_set.prepare_mix_input:
        prepare["mix_input"] = {}
        for key,value in prepare_set.prepare_mix_input.items():
            ivalue = value.split(",")
            prepare["mix_input"][key] = ivalue

    if hasattr(prepare_set,"prepare_mix_kpt") and prepare_set.prepare_mix_kpt:
        kpt_tmp = []
        for ivalue in prepare_set.prepare_mix_kpt.split(","): 
            iivalue = ivalue.split()
            if len(iivalue) == 1:
                try:
                    kpt_tmp.append(int(iivalue[0]))
                except:
                    traceback.print_exc()
                    logs(f"ERROR: The {ivalue} is not valid for KPT!")
            elif len(iivalue) == 3:
                try:
                    kpt_tmp.append([int(iivalue[0]),int(iivalue[1]),int(iivalue[2]),0,0,0])
                except:
                    traceback.print_exc()
                    logs(f"ERROR: The {ivalue} is not valid for KPT!")
            elif len(iivalue) == 6:
                try:
                    kpt_tmp.append([int(iivalue[0]),int(iivalue[1]),int(iivalue[2]),float(iivalue[3]),float(iivalue[4]),float(iivalue[5])])
                except:
                    traceback.print_exc()
                    logs(f"ERROR: The {ivalue} is not valid for KPT!")
            else:
                logs(f"ERROR: The {ivalue} is not valid for KPT!")
                        
        if kpt_tmp:    
            prepare["mix_kpt"] = kpt_tmp
        else:
            logs(f"ERROR: The {prepare_set.prepare_mix_kpt} is not valid for KPT!")
            
    # parse pp lib
    if hasattr(prepare_set,"PreparePPLib"):
        if isinstance(prepare_set.PreparePPLib,PPFromDatasets):
            tmp = None
            try:
                #if prepare_set.PreparePPLib.dataset_unrecorded != None and prepare_set.PreparePPLib.dataset_unrecorded.strip() != "":
                #    url = prepare_set.PreparePPLib.dataset_unrecorded.strip()
                #else:
                url = comm_class_exampleSource.GetDatasetAddress(prepare_set.PreparePPLib.dataset,dataset="pporb")
                package = comm_func.download_url(url, download_path)
                if package == None:
                    logs(f"ERROR: download dataset failed!\n\turl:{url}")
                    logs(f"\tPlease check the dataset!")
                else:
                    comm_func.unpack(package, download_path,filetype="tgz")
                    #remove the package
                    os.remove(package)
                    tmp = True
                    #move the package to work_path
            except:
                traceback.print_exc()
        else:
            tmp = comm_class_exampleSource.parse_source(prepare_set.PreparePPLib,prepare_set.PreparePPLib_local,download_path,prepare_set,logs)
        if tmp:
            pp_path = os.path.join(work_path,"pplib")
            if not os.path.isdir(pp_path):
                os.makedirs(pp_path)
            for ifile in os.listdir(download_path):
                shutil.move(os.path.join(download_path,ifile),os.path.join(pp_path,ifile))
            prepare["pp_path"] = "pplib"
        comm_func.clean_dictorys(download_path)
    elif dataset_work_path and hasattr(prepare_set,"PreparePPLibPath") and getattr(prepare_set,"PreparePPLibPath") and \
        os.path.isdir(os.path.join(dataset_work_path,prepare_set.PreparePPLibPath)):
            prepare["pp_path"] = os.path.join(dataset_work_path,prepare_set.PreparePPLibPath)
    
    #prepare orb lib
    if hasattr(prepare_set,"PreparePPLib"):
        if isinstance(prepare_set.PrepareOrbLib,OrbFromDatasets):
            tmp = None
            try:
                #if prepare_set.PrepareOrbLib.dataset_unrecorded != None and prepare_set.PrepareOrbLib.dataset_unrecorded.strip() != "":
                #    url = prepare_set.PrepareOrbLib.dataset_unrecorded.strip()
                #else:
                url = comm_class_exampleSource.GetDatasetAddress(prepare_set.PrepareOrbLib.dataset,dataset="pporb")
                package = comm_func.download_url(url, download_path)
                if package == None:
                    logs(f"ERROR: download dataset failed!\n\turl:{url}")
                    logs(f"\tPlease check the dataset!")
                else:
                    comm_func.unpack(package, download_path,filetype="tgz")
                    #remove the package
                    os.remove(package)
                    tmp = True
            except:
                traceback.print_exc()
        else:
            tmp = comm_class_exampleSource.parse_source(prepare_set.PrepareOrbLib,prepare_set.PrepareOrbLib_local,download_path,prepare_set,logs)
        if tmp:
            #move the package to work_path
            orb_path = os.path.join(work_path,"orblib")
            if not os.path.isdir(orb_path):
                os.makedirs(orb_path)
            for ifile in os.listdir(download_path):
                shutil.move(os.path.join(download_path,ifile),os.path.join(orb_path,ifile))
            prepare["orb_path"] = "orblib"
        comm_func.clean_dictorys(download_path)
    elif dataset_work_path and hasattr(prepare_set,"PrepareOrbLibPath") and getattr(prepare_set,"PrepareOrbLibPath") and \
        os.path.isdir(os.path.join(dataset_work_path,prepare_set.PrepareOrbLibPath)):
            prepare["orb_path"] = os.path.join(dataset_work_path,prepare_set.PrepareOrbLibPath)
                                                          
    return prepare    
    
def construct_input(datas,opts,work_path,download_path,logs):
    # datas is a dict of examples, created by comm_class_exampleSource.read_source
    #read predft
    need_prepare = False
    logs.iprint("read prepare setting ...")
    prepare = {}

    other_sets = parse_prepare(opts,work_path,download_path,logs)
    if other_sets:
        prepare = other_sets
        need_prepare = True
    
    if datas.get("prepare_example"):
        prepare["example_template"] = datas.get("prepare_example")    
        logs.iprint("\texample:",prepare["example_template"])
        need_prepare = True
    
    if datas.get("prepare_extrafile"):
        prepare["extra_files"] = datas.get("prepare_extrafile")
        logs.iprint("\tprepare_extrafile:",prepare["extra_files"])
        need_prepare = True
    
    return need_prepare, prepare
    



import os,sys,json
import traceback
from .result import Result
from . import comm

class ResultAbacus(Result):
    _PARAM_DIC = {}

    
    def __init__(self, path=".", output=None, resultREF=None, resultCriteria=None):
        """
        Initializes an instance of the ResultAbacus class.

        Args:
            path (str, optional): The path of the ABACUS job. Defaults to ".".
            output (str, optional): The screen output file. Defaults to None, and will be found automatically.
            resultREF (str, optional): The path to the resultRef.json file. Defaults to None. If None, resultREF = os.path.join(self.PATH, "resultRef.json").
        """
        super().__init__()
        self.PATH = path   # the path of ABACUS job

        self.INPUTf = os.path.join(self.PATH, 'INPUT')
        self.STRUf = os.path.join(self.PATH, 'STRU')
        self.KPTf = os.path.join(self.PATH, 'KPT')

        self.INPUT = comm.ReadFile(self.INPUTf, warn=True)
        self.STRU = comm.ReadFile(self.STRUf, warn=True)
        self.KPT = comm.ReadFile(self.KPTf, warn=False)
        
        # screen output
        self.OUTPUTf = output if output is not None else comm.FindOutput(self.PATH, keyinfo="Atomic-orbital Based Ab-initio")
        if self.OUTPUTf is None:
            print("WARNING: can not find the output of ABACUS in %s" % self.PATH)
            self.OUTPUT = []
        else:
            self.OUTPUT = comm.ReadFile(self.OUTPUTf, warn=False)

        self.SUFFIX, self.CALCULATION = self.SuffixCalculation(self.INPUT)

        # OUT.XXX/running_xxx.log
        self.LOGf = os.path.join(self.PATH, "OUT.%s/running_%s.log" % (self.SUFFIX, self.CALCULATION))
        self.LOG = comm.ReadFile(self.LOGf, warn=True)
        
        # time.json
        self.TIME = self.ReadTime()

        # read from abacus.json file
        self.JSON = self.ReadJson()

        # resultREF is a json file of a dict, where key is the param name, and value is the value of the param
        # such as : {"energy": -10000.1111111, "force": [[0.1,0.1,0.1],[0.1,0.1,0.1],[0.1,0.1,0.1]]}
        if resultREF is None:
            resultREF = os.path.join(self.PATH, "resultRef.json")
        self.resultREF = {}
        if os.path.isfile(resultREF):
            try:
                self.resultREF = json.load(open(resultREF))
            except:
                traceback.print_exc()
                print("WARNING: can not read resultRef.json")
                self.resultREF = {}
        

    def SuffixCalculation(self,INPUT):
        suffix = "ABACUS"
        calculation = "scf"
        for iline in INPUT:
            sline = iline.split("#")[0].split()
            if len(sline) >= 2 and sline[0].lower() == "suffix":
                suffix = sline[1].strip()
            elif len(sline) >= 2 and sline[0].lower() == "calculation":
                calculation = sline[1].strip()

        return suffix,calculation
    
    def ReadTime(self): 
        if os.path.isfile(os.path.join(self.PATH,"time.json")):
            time_file = os.path.join(self.PATH,"time.json")
        elif os.path.isfile(os.path.join(self.PATH,f"OUT.{self.SUFFIX}","time.json")):
            time_file = os.path.join(self.PATH,f"OUT.{self.SUFFIX}","time.json")
        else:
            time_file = None
        
        if time_file:
            import json
            try:
                return json.load(open(time_file))
            except:
                traceback.print_exc()
                return None
        else:
            return None

    def ReadJson(self):
        return None
        if os.path.isfile(os.path.join(self.PATH,"abacus.json")):
            try:
                jsons = json.load(open(os.path.join(self.PATH,"abacus.json")))
                return jsons
            except:
                print("WARNING: can not read abacus.json")
                traceback.print_exc()
                return None
        return None
    
    def GetTime(self,class_name,func_name):
        # return the cpu_second of the function, and the calls of the function
        # if tunc_name is None, return the cpu_second of the class and the number of functions in the class
        '''
        time.json:
	    {
	        "total": 1.0,
	        "sub": [
	            {
	                "class_name": "wavefunc",
	                "sub": [
	                    {
	                        "name": "evc",
	                        "cpu_second": 0.000318,
	                        "calls": 2,
	                        "cpu_second_per_call": 0.000159,
	                        "cpu_second_per_total": 0.000318
	                    }
	                ]
	            }
	        ]
	    }
        '''
        if self.TIME == None:
            return (None,None)
        if class_name == "total":
            return (self.TIME.get("total",None),1)
        if "sub" not in self.TIME:
            return (None,None)
        for sub in self.TIME["sub"]:
            if sub.get("class_name",None) == class_name:
                if "sub" not in sub:
                    return (None,None)
                if func_name == None:
                    total_time = 0.0
                    for subsub in sub["sub"]:
                        total_time += subsub.get("cpu_second",0.0)
                    return (total_time,len(sub["sub"]))
                else:
                    for subsub in sub["sub"]:
                        if subsub.get("name",None) == func_name:
                            return (subsub.get("cpu_second",None),subsub.get("calls",None))
                break
        return (None,None)



import os
from dflow.python import upload_packages
root_path = os.path.split(os.path.split(os.path.abspath(__file__))[0])[0]
upload_packages.append(os.path.join(root_path,"abacustest"))

STRU1 = '''ATOMIC_SPECIES
Si 1.000 Si.upf  #Element, Mass, Pseudopotential

NUMERICAL_ORBITAL
Si.orb

LATTICE_CONSTANT
10.2                    #Lattice constant

LATTICE_VECTORS
0.5 0.5 0.0             #Lattice vector 1
0.5 0.0 0.5             #Lattice vector 2
0.0 0.5 0.5             #Lattice vector 3

ATOMIC_POSITIONS
Cartesian               #Cartesian(Unit is LATTICE_CONSTANT)
Si                      #Name of element
0.0                     #Magnetic for this element.
2                       #Number of atoms
0.00 0.00 0.00 0 0 0    #x,y,z, move_x, move_y, move_z
0.25 0.25 0.25 1 1 1
'''

STRU2 = '''ATOMIC_SPECIES
Si 1.000 pplib/Si.upf1  #Element, Mass, Pseudopotential

NUMERICAL_ORBITAL
orblib/Si.orb1

LATTICE_CONSTANT
5.1                    #Lattice constant

LATTICE_VECTORS
0.5 0.5 0.0             #Lattice vector 1
0.5 0.0 0.5             #Lattice vector 2
0.0 0.5 0.5             #Lattice vector 3

ATOMIC_POSITIONS
Cartesian               #Cartesian(Unit is LATTICE_CONSTANT)
Si                      #Name of element
0.0                     #Magnetic for this element.
2                       #Number of atoms
0.00 0.00 0.00 0 0 0    #x,y,z, move_x, move_y, move_z
0.25 0.25 0.25 1 1 1
'''

INPUT1 = """INPUT_PARAMETERS
ecutwfc                 50
basis_type              lcao
"""

INPUT2 = """INPUT_PARAMETERS
basis_type              pw
"""

KPT1 = """K_POINTS
0
Gamma
4 4 4 0 0 0
"""

KPT2 = """K_POINTS
0
Gamma
2 2 2 0 0 0
"""


import os


def show_skills(param=None):
    main_dir = os.path.dirname(os.path.abspath(__file__))
    skills_path = os.path.join(main_dir, "skills")

    if not os.path.isdir(skills_path):
        print(f"ERROR: skills directory not found at {skills_path}")
        return

    print(f"abacustest skills path: {skills_path}")
    print("\nThis is the abacustest skills path. You can add it to the skills directory")
    print("of openclaw/opencode/clawcode and other AI agents. After that, your AI agent")
    print("will be able to correctly handle ABACUS tasks and use abacustest.")


def main():
    import argparse
    parser = argparse.ArgumentParser(description="abacustest skills")
    param = parser.parse_args()
    show_skills(param)


if __name__ == "__main__":
    main()